# Copyright (c) 2014 Rackspace
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os

from oslo_log import log as logging

from solumclient import client as solum_client
from solumclient.common import cliutils
from solumclient.common import exc


class CommandsBase(object):
    """Base command parsing class."""
    parser = None
    solum = None
    json_output = False
    verify = True

    def __init__(self, parser):
        self.parser = parser

        self._get_global_flags()

        try:
            self._get_auth_flags()
        except exc.CommandError as ce:
            print(self.__doc__)
            print("ERROR: %s" % ce.message)
            return

        self.parser.add_argument('action',
                                 default='help',
                                 help='Action to perform on resource')

        parsed, _ = self.parser.parse_known_args()
        action = vars(parsed).get('action')

        client_args = vars(parsed)
        client_args.pop('insecure', None)
        client_args['verify'] = self.verify
        if 'os_auth_token' in client_args:
            del client_args['os_auth_token']

        self.client = solum_client.get_client(parsed.solum_api_version,
                                              **client_args)

        if action in self._actions:
            try:
                return self._actions[action]()
            except exc.CommandError as ce:
                print("ERROR: %s" % ce.message)
        else:
            print(self.__doc__)

    def _get_auth_flags(self):
        self.parser.add_argument('--os-username',
                                 default=env('OS_USERNAME'),
                                 help='Defaults to env[OS_USERNAME]')

        self.parser.add_argument('--os-password',
                                 default=env('OS_PASSWORD'),
                                 help='Defaults to env[OS_PASSWORD]')

        self.parser.add_argument('--os-tenant-name',
                                 default=env('OS_TENANT_NAME'),
                                 help='Defaults to env[OS_TENANT_NAME]')

        self.parser.add_argument('--os-tenant-id',
                                 default=env('OS_TENANT_ID'),
                                 help='Defaults to env[OS_TENANT_ID]')

        self.parser.add_argument('--os-project-name',
                                 default=env('OS_PROJECT_NAME'),
                                 help='Defaults to env[OS_PROJECT_NAME]')

        self.parser.add_argument('--os-project-id',
                                 default=env('OS_PROJECT_ID'),
                                 help='Defaults to env[OS_PROJECT_ID]')

        self.parser.add_argument('--os-project-domain-name',
                                 default=env('OS_PROJECT_DOMAIN_NAME'),
                                 help='Defaults to '
                                      'env[OS_PROJECT_DOMAIN_NAME]')

        self.parser.add_argument('--os-project-domain-id',
                                 default=env('OS_PROJECT_DOMAIN_ID'),
                                 help='Defaults to '
                                      'env[OS_PROJECT_DOMAIN_NAME]')

        self.parser.add_argument('--os-user-domain-name',
                                 default=env('OS_USER_DOMAIN_NAME'),
                                 help='Defaults to env[OS_USER_DOMAIN_NAME]')

        self.parser.add_argument('--os-user-domain-id',
                                 default=env('OS_USER_DOMAIN_ID'),
                                 help='Defaults to env[OS_USER_DOMAIN_NAME]')

        self.parser.add_argument('--os-region-name',
                                 default=env('OS_REGION_NAME'),
                                 help='Defaults to env[OS_REGION_NAME]')

        self.parser.add_argument('--os-auth-url',
                                 default=env('OS_AUTH_URL'),
                                 help='Defaults to env[OS_AUTH_URL]')

        self.parser.add_argument('--os-auth-token',
                                 default=env('OS_AUTH_TOKEN'),
                                 help='Defaults to env[OS_AUTH_TOKEN]')

        self.parser.add_argument('--solum-url',
                                 default=env('SOLUM_URL'),
                                 help='Defaults to env[SOLUM_URL]')

        api_version = env('SOLUM_API_VERSION', default='1')
        self.parser.add_argument('--solum-api-version',
                                 default=api_version,
                                 help='Defaults to env[SOLUM_API_VERSION] '
                                      'or 1')

        parsed, _ = self.parser.parse_known_args()

        client_args = vars(parsed)

        if not parsed.os_auth_token:
            # Remove arguments that are not to be passed to the client in this
            # case.
            del client_args['os_auth_token']

            if not parsed.os_username:
                raise exc.CommandError("You must provide a username via "
                                       "either --os-username or via "
                                       "env[OS_USERNAME]")

            if not parsed.os_password:
                raise exc.CommandError("You must provide a password via "
                                       "either --os-password or via "
                                       "env[OS_PASSWORD]")

            if not (parsed.os_tenant_name or parsed.os_project_name):
                raise exc.CommandError("You must provide a tenant_name via "
                                       "either --os-tenant-name or via "
                                       "env[OS_TENANT_NAME]")

            if not parsed.os_auth_url:
                raise exc.CommandError("You must provide an auth url via "
                                       "either --os-auth-url or via "
                                       "env[OS_AUTH_URL]")

    @property
    def _actions(self):
        """Action handler."""
        return dict((attr, getattr(self, attr))
                    for attr in dir(self)
                    if not attr.startswith('_')
                    and callable(getattr(self, attr)))

    def _get_global_flags(self):
        """Get global flags."""
        # Good location to add_argument() global options like --verbose
        self.parser.add_argument('--json',
                                 action='store_true',
                                 help='JSON formatted output')
        self.parser.add_argument('-k', '--insecure',
                                 action='store_true',
                                 help='Explicitly allow the client to perform'
                                      ' \"insecure SSL\" (https) requests.'
                                      ' The server\'s certificate will not be'
                                      ' verified against any certificate'
                                      ' authorities. This option should be'
                                      ' used with caution.')
        self.parser.add_argument('-d', '--debug',
                                 action='store_true',
                                 help='Print out request and response '
                                      'details.')
        args, _ = self.parser.parse_known_args()
        if args.json:
            self.json_output = True
        if args.insecure:
            self.verify = False
        if args.debug:
            logging.basicConfig(
                format="%(levelname)s (%(module)s) %(message)s",
                level=logging.DEBUG)
            logging.getLogger('iso8601').setLevel(logging.WARNING)
            urllibpool = 'urllib3.connectionpool'
            logging.getLogger(urllibpool).setLevel(logging.WARNING)

    def _sanitized_fields(self, fields):
        def allowed(field):
            if field.startswith('_'):
                return False
            if field == 'manager':
                return False
            if field == 'artifacts':
                return False
            return True
        return [f for f in fields
                if allowed(f)]

    def _print_dict(self, obj, fields, dict_property="Property", wrap=0):
        fields = self._sanitized_fields(fields)
        try:
            # datsun180b: I have no idea why, but following a PATCH
            # app resources need to have this evaluated once or else
            # the subset assignment below fails.
            obj.attrs
        except TypeError:
            pass
        except AttributeError:
            pass
        subset = dict([(f, getattr(obj, f, '')) for f in fields])
        if self.json_output:
            print(json.dumps(subset, indent=2, sort_keys=True))
        else:
            cliutils.print_dict(subset, dict_property, wrap)

    def _print_list(self, objs, fields, formatters=None, sortby_index=0,
                    mixed_case_fields=None, field_labels=None):
        fields = self._sanitized_fields(fields)
        if self.json_output:
            subsets = [dict([(f, getattr(obj, f, '')) for f in fields])
                       for obj in objs]
            print(json.dumps(subsets, indent=2, sort_keys=True))
        else:
            cliutils.print_list(objs, fields, formatters, sortby_index,
                                mixed_case_fields, field_labels)


class NoSubCommands(CommandsBase):
    """Command parsing class that lacks an 'action'."""
    parser = None
    solum = None
    json_output = False
    verify = True

    def __init__(self, parser):
        self.parser = parser

        self._get_global_flags()

        try:
            self._get_auth_flags()
        except exc.CommandError as ce:
            print(self.__doc__)
            print("ERROR: %s" % ce.message)
            return

        parsed, _ = self.parser.parse_known_args()

        client_args = vars(parsed)
        client_args.pop('insecure', None)
        client_args['verify'] = self.verify
        if 'os_auth_token' in client_args:
            del client_args['os_auth_token']

        self.client = solum_client.get_client(parsed.solum_api_version,
                                              **client_args)

        return self.info()

    def info(self):
        pass


def env(*vars, **kwargs):
    """Search for the first defined of possibly many env vars

    Returns the first environment variable defined in vars, or
    returns the default defined in kwargs.
    """
    for v in vars:
        value = os.environ.get(v, None)
        if value:
            return value
    return kwargs.get('default', '')


def filter_ready_lps(lp_list):
    filtered_list = []
    for lp in lp_list:
        if lp.status == 'READY':
            filtered_list.append(lp)

    return filtered_list



# Copyright (c) 2014 Rackspace
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Initial M1 Solum CLI commands implemented (but not REST communications):
* plan create --repo="repo_url" [--build=no] plan_name
* plan delete plan_name
* plan list
* plan show plan_id
* languagepack create <NAME> <GIT_REPO_URL>
* languagepack list
* languagepack show <LP_ID>
* languagepack delete <LP_ID>
* languagepack logs <LP_ID>
* component list


Notes:
* This code is expected to be replaced by the OpenStack Client (OSC) when
    it has progressed a little bit farther as described at:
    https://wiki.openstack.org/wiki/Solum/CLI
* Internationalization will not be added in M1 since this is a prototype
"""

import argparse
import copy
import json
import re
import sys

import httplib2
import jsonschema
from keystoneclient.v2_0 import client as keystoneclient

import solumclient
from solumclient.common.apiclient import exceptions
from solumclient.common import cli_utils
from solumclient.common import exc
from solumclient.common import github
from solumclient.common import yamlutils
from solumclient import config


def name_error_message(name_type):
    return name_type + (" must be 1-100 characters long, only "
                        "contain a-z,0-9,-,_ and start with "
                        "an alphabet character.")


def name_is_valid(string):
    if not string or not string[0].isalpha():
        return False

    try:
        re.match(r'^([a-z0-9-_]{1,100})$', string).group(0)
    except AttributeError:
        return False
    return True


def ValidName(string):
    if not name_is_valid(string):
        raise AttributeError(name_error_message("Names"))
    return string


def lpname_is_valid(string):
    return name_is_valid(string)


def ValidLPName(string):
    if not lpname_is_valid(string):
        raise AttributeError(name_error_message("LP names"))
    return string


def ValidPort(string):
    try:
        port_val = int(string)
        if 1 <= port_val <= 65535:
            return port_val
        else:
            raise ValueError
    except ValueError:
        raise AttributeError("The port should be an integer between 1 and "
                             "65535")


def transform_git_url(git_url, private):
    # try to use a correct git uri
    pt = re.compile(r'github\.com[:/](.+?)/(.+?)($|/.*$|\.git$|\.git/.*$)')
    match = pt.search(git_url)
    if match:
        user_org_name = match.group(1)
        repo = match.group(2)
        if private:
            right_uri = 'git@github.com:%s/%s.git' % (user_org_name, repo)
        else:
            right_uri = 'https://github.com/%s/%s.git' % (user_org_name, repo)
        return right_uri
    else:
        msg = "Provide the git uri in the following format: "
        if not private:
            msg = msg + "https://github.com/<USER>/<REPO>.git"
        else:
            msg = msg + "git@github.com:<USER>/<REPO>.git"
        raise exc.CommandError(message=msg)


def show_public_keys(artifacts):
    public_keys = {}
    if artifacts:
        for arti in artifacts:
            if arti.content and ('public_key' in arti.content):
                public_keys.update(
                    {arti.content['href']: arti.content['public_key']})
    if public_keys:
        print('Important:')
        print('  Solum has generated and uploaded SSH keypair for your ' +
              'private github repository/ies.')
        print('  You may need to add these public SSH keys as github ' +
              'deploy keys, if they were not uploaded successfully.')
        print('  This enables solum to securely ' +
              'clone/pull your private repository/ies.')
        print('  More details on github deploy keys: ' +
              'https://developer.github.com/guides/' +
              'managing-deploy-keys/#deploy-keys\n')
        for href, pub_key in public_keys.items():
            print('%s :\n  %s' % (href, pub_key))


class PlanCommands(cli_utils.CommandsBase):
    """Commands for working with plans.

Available commands:

    solum plan list
        Print an index of all available plans.

    solum plan show <PLAN>
        Print details about a plan.

    solum plan create <PLANFILE> [--param-file <PARAMFILE>]
        Register a plan with Solum.

    solum plan delete <PLAN>
        Destroy a plan. Plans with dependent assemblies cannot be deleted.
    """

    def create(self):
        """Create a plan."""
        self.parser.add_argument('plan_file',
                                 help="A yaml file that defines a plan,"
                                      " check out solum repo for examples")
        self.parser.add_argument('--param-file',
                                 dest='param_file',
                                 help="A yaml file containing custom"
                                      " parameters to be used in the"
                                      " application, check out solum repo for"
                                      " examples")

        self.parser._names['plan_file'] = 'plan file'
        args = self.parser.parse_args()
        try:
            with open(args.plan_file) as definition_file:
                plan_definition = definition_file.read()
                definition = yamlutils.load(plan_definition)
        except IOError:
            message = "Could not open plan file %s." % args.plan_file
            raise exc.CommandError(message=message)
        except ValueError:
            message = ("Plan file %s was not a valid YAML mapping." %
                       args.plan_file)
            raise exc.CommandError(message=message)

        if args.param_file:
            try:
                with open(args.param_file) as param_f:
                    param_definition = param_f.read()
                definition['parameters'] = yamlutils.load(param_definition)
            except IOError:
                message = "Could not open param file %s." % args.param_file
                raise exc.CommandError(message=message)
            except ValueError:
                message = ("Param file %s was not a valid YAML mapping." %
                           args.param_file)
                raise exc.CommandError(message=message)
        plan = self.client.plans.create(yamlutils.dump(definition))
        fields = ['uuid', 'name', 'description', 'uri', 'artifacts']
        artifacts = copy.deepcopy(vars(plan).get('artifacts'))
        self._print_dict(plan, fields, wrap=72)
        show_public_keys(artifacts)

    def delete(self):
        """Delete a plan."""
        self.parser.add_argument('plan_uuid',
                                 help="Tenant/project-wide unique "
                                 "plan uuid or name")
        self.parser._names['plan_uuid'] = 'plan'
        args = self.parser.parse_args()
        plan = self.client.plans.find(name_or_id=args.plan_uuid)
        self.client.plans.delete(plan_id=str(plan.uuid))

    def show(self):
        """Show a plan's resource."""
        self.parser.add_argument('plan_uuid',
                                 help="Plan uuid or name")
        self.parser._names['plan_uuid'] = 'plan'
        args = self.parser.parse_args()
        plan = self.client.plans.find(name_or_id=args.plan_uuid)
        fields = ['uuid', 'name', 'description', 'uri', 'artifacts']
        artifacts = copy.deepcopy(vars(plan).get('artifacts'))
        self._print_dict(plan, fields, wrap=72)
        show_public_keys(artifacts)

    def list(self):
        """List all plans."""
        fields = ['uuid', 'name', 'description']
        plans = self.client.plans.list()
        self._print_list(plans, fields)


class ComponentCommands(cli_utils.CommandsBase):
    """Commands for working with components.

Available commands:

    solum component list
        Print an index of all available components.

    solum component show <UUID>
        Print details about a component.

    """

    def show(self):
        """Show a component's resource."""
        self.parser.add_argument('component_uuid',
                                 help="Component uuid or name")
        self.parser._names['component_uuid'] = 'component'
        args = self.parser.parse_args()
        component = self.client.components.find(name_or_id=args.component_uuid)
        fields = ['uuid', 'name', 'description', 'uri', 'assembly_uuid']
        self._print_dict(component, fields, wrap=72)

    def list(self):
        """List all components."""
        fields = ['uuid', 'name', 'description', 'assembly_uuid']
        components = self.client.components.list()
        self._print_list(components, fields)


class PipelineCommands(cli_utils.CommandsBase):
    """Commands for working with pipelines.

Available commands:

    solum pipeline list
        Print an index of all available pipelines.

    solum pipeline show <PIPELINE>
        Print details about a pipeline.

    solum pipeline create <PLAN_URI> <WORKBOOK_NAME> <NAME>
        Create a pipeline from a given workbook and registered plan.

    solum pipeline delete <PIPELINE>
        Destroy a pipeline.
    """

    def create(self):
        """Create a pipeline."""
        self.parser.add_argument('plan_uri',
                                 help="Tenant/project-wide unique "
                                 "plan (uri/uuid or name)")
        self.parser.add_argument('workbook_name',
                                 help="Workbook name")
        self.parser.add_argument('name',
                                 type=ValidName,
                                 help="Pipeline name")
        self.parser._names['plan_uri'] = 'plan URI'
        self.parser._names['workbook_name'] = 'workbook'
        args = self.parser.parse_args()
        plan_uri = args.plan_uri
        if '/' not in plan_uri:
            # might be a plan uuid/name
            # let's try and be helpful and get the real plan_uri.
            plan = self.client.plans.find(name_or_id=args.plan_uri)
            plan_uri = plan.uri
            print('Note: using plan_uri=%s' % plan_uri)

        pipeline = self.client.pipelines.create(
            name=args.name,
            plan_uri=plan_uri,
            workbook_name=args.workbook_name)
        fields = ['uuid', 'name', 'description',
                  'trigger_uri']
        self._print_dict(pipeline, fields, wrap=72)

    def delete(self):
        """Delete an pipeline."""
        self.parser.add_argument('pipeline_uuid',
                                 help="Pipeline uuid or name")
        self.parser._names['pipeline_uuid'] = 'pipeline'
        args = self.parser.parse_args()
        pipeline = self.client.pipelines.find(name_or_id=args.pipeline_uuid)
        self.client.pipelines.delete(pipeline_id=str(pipeline.uuid))

    def list(self):
        """List all pipelines."""
        fields = ['uuid', 'name', 'description']
        pipelines = self.client.pipelines.list()
        self._print_list(pipelines, fields)

    def show(self):
        """Show a pipeline's resource."""
        self.parser.add_argument('pipeline_uuid',
                                 help="Pipeline uuid or name")
        self.parser._names['pipeline_uuid'] = 'pipeline'
        args = self.parser.parse_args()
        pipelines = self.client.pipelines.find(name_or_id=args.pipeline_uuid)
        fields = ['uuid', 'name', 'description',
                  'trigger_uri', 'workbook_name', 'last_execution']
        self._print_dict(pipelines, fields, wrap=72)


class LanguagePackCommands(cli_utils.CommandsBase):
    """Commands for working with language packs.

Available commands:

    solum lp create <NAME> <GIT_REPO_URL> [--param-file <param-file-name>]
        Create a new language pack from a git repo.
        You can pass an optional parameter file containing a dictionary.
        One use of the parameter file is to pass in credentials for
        infrastructure that should be used to build the lp.

    solum lp list
        Print and index of all available language packs.

    solum lp show <NAME|UUID>
        Print the details of a language pack.

    solum lp delete <NAME|UUID>
        Destroy a language pack.

    solum lp logs <NAME|UUID>
        Show logs for a language pack.
    """

    def create(self):
        """Create a language pack."""
        self.parser.add_argument('--param-file',
                                 dest='paramfile',
                                 help="Local parameter file location")
        self.parser.add_argument('name',
                                 type=ValidLPName,
                                 help="Language pack name.")
        self.parser.add_argument('--name',
                                 type=ValidLPName,
                                 dest='name',
                                 help="Language pack name.")
        self.parser.add_argument('git_url',
                                 help=("Github url of custom "
                                       "language pack repository."))
        self.parser.add_argument('--git_url',
                                 dest='git_url',
                                 help=("Github url of custom "
                                       "language pack repository."))
        self.parser.add_argument('--lp_metadata',
                                 help="Language pack metadata file.")
        self.parser._names['git_url'] = 'repo URL'
        args = self.parser.parse_args()

        param_data = {}
        if args.paramfile is not None:
            try:
                with open(args.paramfile, 'r') as inf:
                    param_data = yamlutils.load(inf.read())
            except Exception as exp:
                raise exc.CommandException(str(exp))

        lp_metadata = None

        if args.lp_metadata:
            with open(args.lp_metadata) as lang_pack_metadata:
                try:
                    lp_metadata = json.dumps(json.load(lang_pack_metadata))
                except ValueError as excp:
                    message = ("Malformed metadata file: %s" % str(excp))
                    raise exc.CommandError(message=message)

        languagepack = {}
        try:
            languagepack = self.client.languagepacks.create(
                name=args.name, source_uri=args.git_url,
                lp_metadata=lp_metadata,
                lp_params=param_data)
        except exceptions.Conflict as conflict:
            message = ("%s" % conflict.message)
            raise exc.CommandError(message=message)

        fields = ['uuid', 'name', 'description', 'status', 'source_uri']
        self._print_dict(languagepack, fields, wrap=72)

    def delete(self):
        """Delete a language pack."""
        self.parser.add_argument('lp_id',
                                 help="Language pack id")
        self.parser._names['lp_id'] = 'languagepack'
        args = self.parser.parse_args()
        self.client.languagepacks.delete(lp_id=args.lp_id)

    def list(self):
        """List all language packs."""
        fields = ['uuid', 'name', 'description', 'status', 'source_uri']
        languagepacks = self.client.languagepacks.list()
        self._print_list(languagepacks, fields)

    def show(self):
        """Get a language pack."""
        self.parser.add_argument('lp_id',
                                 help="Language pack id")
        self.parser._names['lp_id'] = 'languagepack'
        args = self.parser.parse_args()
        languagepack = self.client.languagepacks.find(name_or_id=args.lp_id)
        fields = ['uuid', 'name', 'description', 'status', 'source_uri']
        self._print_dict(languagepack, fields, wrap=72)

    def logs(self):
        """Get language pack Logs."""
        self.parser.add_argument('lp_id',
                                 help="languagepack uuid or name")
        args = self.parser.parse_args()
        loglist = self.client.languagepacks.logs(lp_id=str(args.lp_id))

        fields = ["resource_uuid", "created_at"]
        for log in loglist:
            strategy_info = json.loads(log.strategy_info)
            if log.strategy == 'local':
                if 'local_storage' not in fields:
                    fields.append('local_storage')
                log.local_storage = log.location
            elif log.strategy == 'swift':
                if 'swift_container' not in fields:
                    fields.append('swift_container')
                if 'swift_path' not in fields:
                    fields.append('swift_path')
                log.swift_container = strategy_info['container']
                log.swift_path = log.location
            else:
                if 'location' not in fields:
                    fields.append('location')

        self._print_list(loglist, fields)


class AppCommands(cli_utils.CommandsBase):
    """Commands for working with actual applications.

Available commands:
    solum app list
        Print an index of all deployed applications.

    solum app show <NAME|ID>
        Print detailed information about one application.

    solum app create [--app-file <AppFile>] [--git-url <GIT_URL>]
                     [--lp <LANGUAGEPACK>]
                     [--param-file <PARAMFILE>]
                     [--setup-trigger]
                     [--port <PORT>]
                     [--private-repo]
                     [--no-languagepack]
                     [--trigger-workflow <CUSTOM-WORKFLOW>]
                      <CUSTOM-WORKFLOW>=(unittest | build | unittest+build)
                      Without the --trigger-workflow flag,
                      the workflow unittest+build+deploy is triggered
                      (this is the default workflow)

        Register a new application with Solum.

    solum app deploy <NAME|ID> --du-id <du-id>
        Deploy an application, building any applicable artifacts first.
        du-id is optional flag. It can be used to pass in id of a previously
        created deployment unit. If passed, this command will deploy that
        du instead of building one first.

    solum app delete <NAME|ID>
        Delete an application and all related artifacts.

    solum app scale <NAME|ID> <scaling target>

    solum app logs <NAME|UUID> [--wf-id <wf-id>]
        Show the logs of an application for all the workflows.
        wf-id is optional flag which can be used to pass in id of one of
        the existing workflows. If provided, the logs only for that workflow
        are displayed.
    """

    def _validate_app_file(self, app_data):
        if ('workflow_config' in app_data and
                app_data.get('workflow_config') is None):
            msg = "Workflow config cannot be empty"
            raise exc.CommandException(message=msg)
        if ('trigger_actions' in app_data and
                app_data.get('trigger_actions') is None):
            msg = "Trigger actions cannot be empty"
            raise exc.CommandException(message=msg)

        error_message = ("Application name must be 1-100 characters long, "
                         "only contain a-z,0-9,-,_ and start with an alphabet "
                         "character.")

        if app_data.get('name') is not None:
            if not name_is_valid(app_data.get('name')):
                raise exc.CommandError(message=error_message)

        if 'repo_token' not in app_data:
            app_data['repo_token'] = ''

    def _get_and_validate_app_name(self, app_data, args):
        # Check the appfile-supplied name first.
        error_message = ("Application name must be 1-100 characters long, "
                         "only contain a-z,0-9,-,_ and start with an alphabet "
                         "character.")
        app_name = ''
        if args.name:
            app_name = args.name
        elif app_data.get('name') is None:
            while True:
                app_name = input("Please name the application.\n> ")
                if name_is_valid(app_name):
                    break
                print(error_message)
        else:
            app_name = app_data.get('name')

        if not name_is_valid(app_name):
            raise exc.CommandError(message=error_message)

        return app_name

    def _get_and_validate_languagepack(self, app_data, args):
        # Check for the language pack. Check args first, then appfile.
        # If it's neither of those places, prompt for it and update the
        # app data.

        if args.no_languagepack:
            app_data['languagepack'] = "False"
            return

        languagepack = None
        if args.languagepack is not None:
            languagepack = args.languagepack
        elif app_data.get('languagepack') is not None:
            languagepack = app_data.get('languagepack')
        # check if given languagepack exists or not
        if languagepack:
            try:
                lp = self.client.languagepacks.find(name_or_id=languagepack)
            except Exception:
                raise exc.CommandError(
                    "Languagepack '%s' not found." % languagepack)
            if lp is None or lp.status != 'READY':
                raise exc.CommandError("No languagepack in READY state. "
                                       "Create a languagepack first.")
            app_data['languagepack'] = languagepack
        else:
            languagepacks = self.client.languagepacks.list()
            filtered_list = cli_utils.filter_ready_lps(languagepacks)

            if len(filtered_list) > 0:
                lpnames = [lang_pack.name for lang_pack in filtered_list]
                lp_uuids = [lang_pack.uuid for lang_pack in filtered_list]
                fields = ['uuid', 'name', 'description',
                          'status', 'source_uri']
                self._print_list(filtered_list, fields)
                languagepack = input("Please choose a languagepack "
                                     "from the above list.\n> ")
                while languagepack not in lpnames + lp_uuids:
                    languagepack = input("You must choose one of "
                                         "the named  language "
                                         "packs.\n> ")
                app_data['languagepack'] = languagepack
            else:
                raise exc.CommandError("No languagepack in READY state. "
                                       "Create a languagepack first.")

    def _get_app_repo_details(self, app_data, args):
        def read_private_sshkey(sshkey_file):
            private_sshkey = ''
            try:
                with open(sshkey_file, 'r') as inf:
                    private_sshkey = inf.read()
            except Exception as exp:
                raise exc.CommandException(str(exp))
            return private_sshkey

        git_rev = 'master'
        git_url = None
        if (app_data.get('source') is not None and
                app_data.get('source').get('repository') is not None):
            git_url = app_data['source']['repository']
        # Commandline flag overrides stuff in the app-file
        if args.git_url is not None:
            git_url = args.git_url
        # Take input from user
        elif (app_data.get('source') is None or
                app_data['source'].get('repository') is None or
                app_data['source']['repository'] == ''):
            git_url = input("Please specify a git repository URL "
                            "for your application.\n> ")
            git_rev_i = input("Please specify revision"
                              "(default is master).\n> ")
            if git_rev_i == '':
                git_rev = 'master'
            else:
                git_rev = git_rev_i

        assert(git_url is not None)
        assert(git_rev is not None)

        # check if given repo is a private repository
        is_private = (args.private_repo or
                      app_data['source'].get('private', False))

        git_url = transform_git_url(git_url, is_private)

        private_sshkey = app_data['source'].get('private_ssh_key', '')
        if is_private and not private_sshkey:
            sshkey_file = input("Please specify private sshkey file "
                                "full path: ")
            sshkey_file = sshkey_file.strip()
            private_sshkey = read_private_sshkey(sshkey_file)

        if is_private and private_sshkey == '':
            msg = "Must provide private sshkey for private repositories."
            raise exc.CommandError(message=msg)

        git_src = dict()
        git_src['private'] = is_private
        git_src['private_ssh_key'] = private_sshkey
        git_src['repository'] = git_url
        git_src['revision'] = git_rev
        app_data['source'] = git_src

    def _get_run_command(self, app_data, args):
        run_cmd = None

        if args.run_cmd is not None:
            run_cmd = args.run_cmd
        elif (app_data.get('workflow_config') is None or
                app_data['workflow_config'].get('run_cmd') == '' or
                app_data['workflow_config'].get('run_cmd') is None):
            run_cmd = input("Please specify start/run command for "
                            "your application.\n> ")

        if app_data.get('workflow_config') is None:
            run_cmd_dict = dict()
            run_cmd_dict['run_cmd'] = run_cmd
            app_data['workflow_config'] = run_cmd_dict
        elif (app_data['workflow_config'].get('run_cmd') == '' or
              app_data['workflow_config'].get('run_cmd') is None):
            app_data['workflow_config']['run_cmd'] = run_cmd

    def _get_unittest_command(self, app_data, args):
        unittest_cmd = None

        if args.unittest_cmd is not None:
            unittest_cmd = args.unittest_cmd
            if app_data.get('workflow_config') is None:
                unittest_cmd_dict = dict()
                unittest_cmd_dict['test_cmd'] = unittest_cmd
                app_data['workflow_config'] = unittest_cmd_dict
            elif (app_data['workflow_config'].get('test_cmd') == '' or
                  app_data['workflow_config'].get('test_cmd') is None):
                app_data['workflow_config']['test_cmd'] = unittest_cmd

    def _get_port(self, app_data, args):
        port_list = []

        if (app_data.get('ports') is None or
                app_data['ports'] == '' or app_data['ports'] == [None]):
            if args.port:
                port_list.append(int(args.port))
            else:
                print("Using 80 as the app's default listening port")
                port_list.append(int(80))

            app_data['ports'] = port_list

    def _get_parameters(self, app_data, args):
        app_data['parameters'] = {}
        if args.param_file is not None:
            try:
                with open(args.param_file) as param_f:
                    param_def = param_f.read()
                app_data['parameters'] = yamlutils.load(param_def)
            except IOError:
                message = "Could not open param file %s." % args.param_file
                raise exc.CommandError(message=message)
            except ValueError:
                message = ("Param file %s was not YAML." %
                           args.param_file)
                raise exc.CommandError(message=message)

    def _setup_github_trigger(self, app_data, app, args):
        # If a token is supplied, we won't need to generate one.
        provided_token = ''
        obtained_token = ''
        if 'repo_token' in app_data:
            provided_token = app_data['repo_token']

        if args.setup_trigger or args.workflow:
            trigger_uri = vars(app).get('trigger_uri', '')
            if trigger_uri:
                workflow = None
                if args.workflow:
                    workflow = args.workflow.replace('+', ' ').split(' ')
                try:
                    git_url = app_data['source']['repository']

                    gha = github.GitHubAuth(git_url,
                                            repo_token=provided_token)
                    gha.create_webhook(trigger_uri, workflow=workflow)

                    if not provided_token:
                        try:
                            obtained_token = gha.create_repo_token()
                            # Update the app with the generated repo_token
                            to_update = {}
                            to_update['repo_token'] = obtained_token
                            self.client.apps.patch(app_id=app.id, **to_update)
                        except Exception:
                            print("Error in obtaining github token.")
                            return ''

                except github.GitHubException as ghe:
                    raise exc.CommandError(message=str(ghe))

    def create(self):

        parsed, _ = self.parser.parse_known_args()
        config.username = parsed.os_username
        config.password = parsed.os_password
        config.tenant = parsed.os_tenant_name

        self.register()

    def register(self):
        """Register a new app."""
        self.parser.add_argument('--app-file',
                                 dest='appfile',
                                 help="Local appfile location")
        self.parser.add_argument('--name',
                                 dest='name',
                                 type=ValidName,
                                 help="Application name")
        self.parser.add_argument('--languagepack',
                                 help='Language pack')
        self.parser.add_argument('--lp',
                                 dest='languagepack',
                                 help='Language pack')
        self.parser.add_argument('--git-url',
                                 dest='git_url',
                                 help='Source repo')
        self.parser.add_argument('--run-cmd',
                                 dest='run_cmd',
                                 help="Application entry point")
        self.parser.add_argument('--unittest-cmd',
                                 dest='unittest_cmd',
                                 help="Command to execute unit tests")
        self.parser.add_argument('--port',
                                 dest="port",
                                 type=ValidPort,
                                 help="The port your application listens on")
        self.parser.add_argument('--param-file',
                                 dest='param_file',
                                 help="A yaml file containing custom"
                                      " parameters to be used in the"
                                      " application")
        self.parser.add_argument('--setup-trigger',
                                 action='store_true',
                                 dest='setup_trigger',
                                 help="Set up app trigger on git repo")
        self.parser.add_argument('--no-languagepack',
                                 action='store_true',
                                 dest='no_languagepack',
                                 help="Flag to register an app without"
                                      " a languagepack")
        self.parser.add_argument('--private-repo',
                                 action='store_true',
                                 dest='private_repo',
                                 help="Source repo requires authentication.")

        trigger_help = ("Which of stages build, unittest, deploy to trigger "
                        "from git. For example: "
                        "--trigger-workflow=unittest+build+deploy. "
                        "Implies --setup-trigger.")
        self.parser.add_argument('--trigger-workflow',
                                 default='',
                                 dest='workflow',
                                 help=trigger_help)

        args = self.parser.parse_args()
        app_data = None
        if args.appfile is not None:
            try:
                with open(args.appfile, 'r') as inf:
                    app_data = yamlutils.load(inf.read())
                    self._validate_app_file(app_data)
            except Exception as exp:
                raise exc.CommandException(str(exp))
        else:
            app_data = {
                'version': 1,
                'description': 'default app description.',
                'source': {
                    'repository': '',
                    'revision': 'master',
                    'private': False,
                    'private_ssh_key': '',
                    'repo_token': ''
                },
                'workflow_config': {
                    'test_cmd': '',
                    'run_cmd': ''
                },
                'trigger_actions': ['build', 'deploy'],
            }

        app_name = self._get_and_validate_app_name(app_data, args)
        app_data['name'] = app_name

        self._get_and_validate_languagepack(app_data, args)

        self._get_app_repo_details(app_data, args)

        self._get_run_command(app_data, args)

        self._get_unittest_command(app_data, args)

        self._get_port(app_data, args)

        self._get_parameters(app_data, args)

        self._validate_app_data(app_data)

        if args.workflow:
            app_data['trigger_actions'] = (
                args.workflow.replace('+', ' ').split(' '))

        app = self.client.apps.create(**app_data)

        self._setup_github_trigger(app_data, app, args)

        app.trigger = app.trigger_actions
        app.workflow = app.workflow_config

        fields = ['name', 'id', 'created_at', 'description', 'languagepack',
                  'ports', 'source', 'workflow',
                  'trigger_uuid', 'trigger', 'trigger_uri']
        self._print_dict(app, fields, wrap=72)

    def _validate_app_data(self, app_data):
        # app file schema
        schema = {
            "title": "app file schema",
            "type": "object",
            "properties": {
                "version": {
                    "type": "integer"
                },
                "name": {
                    "type": "string"
                },
                "description": {
                    "type": "string"
                },
                "languagepack": {
                    "type": "string"
                },
                "source": {
                    "type": "object",
                    "properties": {
                        "repository": {
                            "type": "string"
                        },
                        "revision": {
                            "type": "string"
                        },
                        "private": {
                            "type": "boolean"
                        },
                        "repo_token": {
                            "type": "string"
                        },
                        "private_ssh_key": {
                            "type": "string"
                        }
                    }
                },
                "trigger_actions": {
                    "type": "array",
                    "items": {
                        "type": "string",
                        "enum": ["unittest", "build", "deploy"]
                    }
                },
                "workflow_config": {
                    "type": "object",
                    "properties": {
                        "test_cmd": {
                            "type": "string"
                        },
                        "run_cmd": {
                            "type": "string"
                        }
                    }
                },
                "ports": {
                    "type": "array",
                    "items": {
                        "type": "integer"
                    }
                },
                "parameters": {
                    "type": "object"
                }
            },
            "required": [
                "version",
                "name",
                "description",
                "languagepack",
                "source",
                "trigger_actions",
                "workflow_config",
                "ports"
            ]
        }
        try:
            jsonschema.validate(app_data, schema)
        except jsonschema.exceptions.ValidationError as exp:
            raise exc.CommandError(message=str(exp))

    def update(self):
        """Update the registration of an existing app."""
        self.parser.add_argument('app')

        self.parser.add_argument('--name', type=str)
        self.parser.add_argument('--desc', type=str)
        self.parser.add_argument('--lp', type=str)
        self.parser.add_argument('--ports', type=str)
        self.parser.add_argument('--source.repo', dest='source_repo', type=str)
        self.parser.add_argument('--source.rev', dest='source_rev', type=str)
        self.parser.add_argument('--test_cmd', type=str)
        self.parser.add_argument('--run_cmd', type=str)
        self.parser.add_argument('--trigger', type=str)

        args = self.parser.parse_args()

        app = self.client.apps.find(name_or_id=args.app)

        to_update = {}
        if args.name:
            to_update['name'] = args.name
        if args.desc:
            to_update['description'] = args.desc
        if args.lp:
            to_update['languagepack'] = args.lp
        if args.ports:
            ports = args.ports.strip('[]').replace(',', ' ')
            ports = [int(p, 10) for p in ports.split(' ') if p]
            to_update['ports'] = ports
        if args.source_repo:
            to_update['source'] = to_update.get('source', {})
            to_update['source']['repository'] = args.source_repo
        if args.source_rev:
            to_update['source'] = to_update.get('source', {})
            to_update['source']['revision'] = args.source_rev
        if args.test_cmd:
            to_update['workflow_config'] = to_update.get('trigger', {})
            to_update['workflow_config']['test_cmd'] = args.test_cmd
        if args.run_cmd:
            to_update['workflow_config'] = to_update.get('trigger', {})
            to_update['workflow_config']['run_cmd'] = args.run_cmd
        if args.trigger:
            trigger = args.trigger.strip('[]').replace(',', ' ').split(' ')
            to_update['trigger_actions'] = trigger

        if not to_update:
            raise exc.CommandException(message="Nothing to update")

        updated_app = self.client.apps.patch(app_id=app.id, **to_update)

        updated_app.trigger = updated_app.trigger_actions
        updated_app.workflow = updated_app.workflow_config

        fields = ['name', 'id', 'created_at', 'description', 'languagepack',
                  'entry_points', 'ports', 'source', 'workflow',
                  'trigger_uuid', 'trigger']
        self._print_dict(updated_app, fields, wrap=72)

    def list(self):
        """List all apps."""
        apps = self.client.apps.list()
        fields = ['name', 'id', 'created_at', 'description', 'languagepack']
        self._print_list(apps, fields)

    def show(self):
        """Show details of one app."""
        self.parser.add_argument('name')
        args = self.parser.parse_args()
        app = self.client.apps.find(name_or_id=args.name)

        fields = ['name', 'id', 'created_at', 'updated_at', 'description',
                  'languagepack', 'entry_points', 'ports', 'source',
                  'trigger_uuid', 'trigger', 'app_url']

        app.trigger = app.trigger_actions
        app.workflow = app.workflow_config
        if app.scale_config:
            if app.scale_config[app.name]['target']:
                app.target_instances = app.scale_config[app.name]['target']
                fields.append('target_instances')

        self._print_dict(app, fields, wrap=72)

        wfs = self.client.workflows.list(app_id=app.id)
        fields = ['wf_id', 'id', 'status']
        print("'%s' workflows and their status:" % args.name)
        self._print_list(wfs, fields)

    def delete(self):
        """Delete an app."""
        self.parser.add_argument('name')
        args = self.parser.parse_args()
        app = self.client.apps.find(name_or_id=args.name)
        self.client.apps.delete(app_id=str(app.id))

    def _create_scaling_workflow(self, actions, app_name_id, tgt):
        app = self.client.apps.find(name_or_id=app_name_id)
        kwargs = {'app_id': app.id, 'actions': actions,
                  'scale_target': tgt}
        wf = self.client.workflows.create(**kwargs)
        fields = ['wf_id', 'app_id', 'actions', 'config',
                  'source', 'id', 'created_at', 'updated_at']
        self._print_dict(wf, fields, wrap=72)

    def _create_workflow(self, actions, app_name_id=''):
        if not app_name_id:
            app = self.client.apps.find(name_or_id=app_name_id)
        else:
            self.parser.add_argument('name')
            args = self.parser.parse_args()
            app = self.client.apps.find(name_or_id=args.name)
        kwargs = {'app_id': app.id, 'actions': actions}
        wf = self.client.workflows.create(**kwargs)
        fields = ['wf_id', 'app_id', 'actions', 'config',
                  'source', 'id', 'created_at', 'updated_at']
        self._print_dict(wf, fields, wrap=72)

    def _create_workflow_for_prebuilt_du(self, actions, app_name_id, du_id):
        app = self.client.apps.find(name_or_id=app_name_id)
        kwargs = {'app_id': app.id, 'actions': actions,
                  'du_id': du_id}
        wf = self.client.workflows.create(**kwargs)
        fields = ['wf_id', 'app_id', 'actions', 'config',
                  'source', 'id', 'created_at', 'updated_at']
        self._print_dict(wf, fields, wrap=72)

    def unittest(self):
        """Create a new workflow for an app."""
        actions = ['unittest']
        self._create_workflow(actions)

    def build(self):
        """Create a new workflow for an app."""
        actions = ['unittest', 'build']
        self._create_workflow(actions)

    def deploy(self):
        """Create a new workflow for an app."""
        self.parser.add_argument("name")
        self.parser.add_argument('--du-id',
                                 dest="du_id",
                                 help="ID of the DU image.")
        args = self.parser.parse_args()
        if args.du_id:
            actions = ['deploy']
            self._create_workflow_for_prebuilt_du(actions, args.name,
                                                  args.du_id)
        else:
            actions = ['unittest', 'build', 'deploy']
            self._create_workflow(actions, args.name)

    def scale(self):
        """Scale the app."""
        self.parser.add_argument('name')
        self.parser.add_argument('target')
        args = self.parser.parse_args()

        target = args.target
        try:
            target = int(target)
        except ValueError:
            msg = "Must provide integer value for scale target."
            raise exc.CommandException(message=msg)

        if target <= 0:
            msg = "Scale target must be greater than zero."
            raise exc.CommandException(message=msg)

        actions = ['scale']
        self._create_scaling_workflow(actions, args.name, target)

    def _display_logs_for_all_workflows(self, app):
        wfs = self.client.workflows.list(app_id=app.id)

        all_logs_list = []
        fields = ["resource_uuid", "created_at"]
        for wf in wfs:
            revision = wf.wf_id
            kwargs = {'app_id': app.id, 'revision_or_id': revision}
            loglist = self.client.workflows.logs(**kwargs)
            for log in loglist:
                all_logs_list.append(log)
                strategy_info = json.loads(log.strategy_info)
                if log.strategy == 'local':
                    if 'local_storage' not in fields:
                        fields.append('local_storage')
                    log.local_storage = log.location
                elif log.strategy == 'swift':
                    if 'swift_container' not in fields:
                        fields.append('swift_container')
                    if 'swift_path' not in fields:
                        fields.append('swift_path')
                    log.swift_container = strategy_info['container']
                    log.swift_path = log.location
                else:
                    if 'location' not in fields:
                        fields.append('location')
        self._print_list(all_logs_list, fields)

    def logs(self):
        """Print a list of all logs belonging to a single app."""

        self.parser.add_argument('name')

        self.parser.add_argument('--wf-id',
                                 dest='wf_id',
                                 help="Workflow ID")

        args = self.parser.parse_args()
        app = self.client.apps.find(name_or_id=args.name)

        if args.wf_id:
            try:
                revision = int(args.wf_id, 10)
            except (ValueError, TypeError):
                revision = args.wf_id
            display_logs_for_single_workflow(self, app, revision)
        else:
            self._display_logs_for_all_workflows(app)


def display_logs_for_single_workflow(self, app, revision):
    kwargs = {'app_id': app.id, 'revision_or_id': revision}
    loglist = self.client.workflows.logs(**kwargs)
    fields = ["resource_uuid", "created_at"]
    for log in loglist:
        strategy_info = json.loads(log.strategy_info)
        if log.strategy == 'local':
            if 'local_storage' not in fields:
                fields.append('local_storage')
            log.local_storage = log.location
        elif log.strategy == 'swift':
            if 'swift_container' not in fields:
                fields.append('swift_container')
            if 'swift_path' not in fields:
                fields.append('swift_path')
            log.swift_container = strategy_info['container']
            log.swift_path = log.location
        else:
            if 'location' not in fields:
                fields.append('location')

    self._print_list(loglist, fields)


class WorkflowCommands(cli_utils.CommandsBase):
    """Commands for working with workflows.

Available commands:

    solum workflow list <APP_NAME|UUID>
        List all application workflows.

    solum workflow show <APP_NAME|UUID> <WORKFLOW_ID|UUID>
        Print the details of a workflow.

    solum workflow logs <APP_NAME|UUID> <WORKFLOW_ID|UUID>
        List all the logs of a given workflow.
    """

    def list(self):
        """Show all of an app's live workflows."""
        self.parser.add_argument('app')
        args = self.parser.parse_args()
        app = self.client.apps.find(name_or_id=args.app)
        wfs = self.client.workflows.list(app_id=app.id)
        fields = ['wf_id', 'id', 'actiONs', 'status',
                  'created_at', 'updated_at']
        self._print_list(wfs, fields)

    def show(self):
        """Show one of an app's live workflows."""
        # Either "solum workflow show <app_id_or_name> <workflow_uuid>
        # Or "solum workflow show <app_id_or_name> <workflow_revision>
        self.parser.add_argument('app')
        self.parser.add_argument('workflow')
        args = self.parser.parse_args()
        revision = args.workflow
        try:
            revision = int(revision, 10)
        except ValueError:
            revision = args.workflow
        app = self.client.apps.find(name_or_id=args.app)

        kwargs = {'app_id': app.id, 'revision_or_id': revision}
        wf = self.client.workflows.find(**kwargs)

        fields = ['wf_id', 'app_id', 'actions', 'config',
                  'source', 'id', 'created_at', 'updated_at', 'status']
        self._print_dict(wf, fields, wrap=72)

    def logs(self):
        """Show one of an app's live workflows logs."""
        self.parser.add_argument('app',
                                 help="App uuid or name")
        self.parser.add_argument('workflow',
                                 help="Workflow id or uuid")
        args = self.parser.parse_args()
        revision = args.workflow
        try:
            revision = int(revision, 10)
        except ValueError:
            revision = args.workflow
        app = self.client.apps.find(name_or_id=args.app)
        display_logs_for_single_workflow(self, app, revision)


class OldAppCommands(cli_utils.CommandsBase):
    """Commands for working with applications.

Available commands:
    solum app list
        Print an index of all deployed applications.

    solum app show <NAME|UUID>
        Print detailed information about one application.

    solum app create [--plan-file <PLANFILE>] [--git-url <GIT_URL>]
                     [--lp <LANGUAGEPACK>] [--run-cmd <RUN_CMD>]
                     [--unittest-cmd <UNITTEST_CMD>]
                     [--name <NAME>] [--port <PORT>]
                     [--param-file <PARAMFILE>]
                     [--desc <DESCRIPTION>]
                     [--setup-trigger]
                     [--private-repo]
                     [--trigger-workflow <WORKFLOW>]
        Register a new application with Solum.

    solum app deploy <NAME|UUID>
        Deploy an application, building any applicable artifacts first.

    solum app logs <NAME|UUID>
        Show the logs of an application for all the deployments.

    solum app delete <NAME|UUID>
        Delete an application and all related artifacts.
"""

    def _get_assemblies_by_plan(self, plan):
        # TODO(datsun180b): Write this.
        return []

    def _validate_plan_file(self, plan_definition):
        if 'artifacts' not in plan_definition:
            raise exc.CommandException(message="Missing artifacts section")
        elif plan_definition['artifacts'] is None:
            raise exc.CommandException(message="Artifacts cannot be empty")
        elif 'content' not in plan_definition['artifacts'][0]:
            raise exc.CommandException(message="Artifact content missing")

        error_message = ("Application name must be 1-100 characters long, "
                         "only contain a-z,0-9,-,_ and start with an alphabet "
                         "character.")

        if plan_definition.get('name') is not None:
            if not name_is_valid(plan_definition.get('name')):
                raise exc.CommandError(message=error_message)

    def list(self):
        """Print a list of all deployed applications."""
        # This is just "plan list".
        # TODO(datsun180b): List each plan and its associated
        # assemblies.
        fields = ['uuid', 'name', 'description']
        plans = self.client.plans.list()
        self._print_list(plans, fields)

    def show(self):
        """Print detailed information about one application."""
        # This is just "plan show <PLAN>".
        # TODO(datsun180b): List the details of the plan, and
        # also the current build state, build number, and running
        # assembly status. We don't have all the pieces for that yet.
        self.parser.add_argument('app',
                                 help="Application name")
        self.parser._names['app'] = 'application'
        args = self.parser.parse_args()
        try:
            plan = self.client.plans.find(name_or_id=args.app)
        except exceptions.NotFound:
            message = "No app named '%s'." % args.app
            raise exceptions.NotFound(message=message)

        # Fetch the most recent app_uri.
        assemblies = self.client.assemblies.list()
        app_uri = ''
        updated = ''
        app_status = 'REGISTERED'
        for a in assemblies:
            plan_uuid = a.plan_uri.split('/')[-1]
            if plan_uuid != plan.uuid:
                continue
            if a.updated_at >= updated:
                updated = a.updated_at
                app_uri = a.application_uri
                app_status = a.status

        plan.application_uri = app_uri
        plan.status = app_status
        fields = ['uuid', 'name', 'description', 'uri', 'artifacts',
                  'trigger_uri', 'application_uri', 'status']
        self._print_dict(plan, fields, wrap=72)
        artifacts = copy.deepcopy(vars(plan).get('artifacts'))
        show_public_keys(artifacts)

    def create(self):
        """Register a new application with Solum."""
        # This is just "plan create" with a little proactive
        # parsing of the planfile.

        self.parser.add_argument('--plan-file',
                                 dest='planfile',
                                 help="Local planfile location")
        self.parser.add_argument('--git-url',
                                 help='Source repo')
        self.parser.add_argument('--languagepack',
                                 help='Language pack')
        self.parser.add_argument('--lp',
                                 dest='languagepack',
                                 help='Language pack')
        self.parser.add_argument('--run-cmd',
                                 help="Application entry point")
        self.parser.add_argument('--unittest-cmd',
                                 help="Command to execute unit tests")
        self.parser.add_argument('--port',
                                 type=ValidPort,
                                 help="The port your application listens on")
        self.parser.add_argument('--name',
                                 type=ValidName,
                                 help="Application name")
        self.parser.add_argument('--desc',
                                 help="Application description")
        self.parser.add_argument('--param-file',
                                 dest='param_file',
                                 help="A yaml file containing custom"
                                      " parameters to be used in the"
                                      " application")
        self.parser.add_argument('--setup-trigger',
                                 action='store_true',
                                 dest='setup_trigger',
                                 help="Set up app trigger on git repo")
        self.parser.add_argument('--private-repo',
                                 action='store_true',
                                 dest='private_repo',
                                 help="Source repo requires authentication.")

        trigger_help = ("Which of stages build, unittest, deploy to trigger "
                        "from git. For example: "
                        "--trigger-workflow=unittest+build+deploy. "
                        "Implies --setup-trigger.")
        self.parser.add_argument('--trigger-workflow',
                                 default='',
                                 dest='workflow',
                                 help=trigger_help)

        args = self.parser.parse_args()

        # Get the plan file. Either get it from args, or supply
        # a skeleton.
        plan_definition = None
        if args.planfile is not None:
            planfile = args.planfile
            try:
                with open(planfile) as definition_file:
                    definition = definition_file.read()
                    plan_definition = yamlutils.load(definition)
                    self._validate_plan_file(plan_definition)
                    if (plan_definition['artifacts'][0].get('language_pack')
                            is not None):
                        lp = plan_definition['artifacts'][0]['language_pack']
                        if lp != 'auto':
                            try:
                                lp1 = (
                                    self.client.languagepacks.find
                                    (name_or_id=lp)
                                )
                            except Exception as e:
                                if type(e).__name__ == 'NotFound':
                                    raise exc.CommandError("Languagepack %s "
                                                           "not registered"
                                                           % lp)
                            filtered_list = cli_utils.filter_ready_lps([lp1])
                            if len(filtered_list) <= 0:
                                raise exc.CommandError("Languagepack %s "
                                                       "not READY" % lp)
                    if plan_definition['artifacts'][0].get('ports') is None:
                        print("No application port specified in plan file.")
                        print("Defaulting to port 80.")
                        plan_definition['artifacts'][0]['ports'] = 80
            except IOError:
                message = "Could not open plan file %s." % planfile
                raise exc.CommandError(message=message)
            except ValueError:
                message = ("Plan file %s was not a valid YAML mapping." %
                           planfile)
                raise exc.CommandError(message=message)
        else:
            plan_definition = {
                'version': 1,
                'artifacts': [{
                    'artifact_type': 'heroku',
                    'name': '',
                    'content': {},
                }]}

        # NOTE: This assumes the plan contains exactly one artifact.

        # Check the planfile-supplied name first.
        error_message = ("Application name must be 1-100 characters long, "
                         "only contain a-z,0-9,-,_ and start with an alphabet "
                         "character.")
        app_name = ''
        if plan_definition.get('name') is not None:
            if not name_is_valid(plan_definition.get('name')):
                raise exc.CommandError(message=error_message)
            app_name = plan_definition.get('name')
        # Check the arguments next.
        elif args.name:
            if name_is_valid(args.name):
                app_name = args.name
        # Just ask.
        else:
            while True:
                app_name = input("Please name the application.\n> ")
                if name_is_valid(app_name):
                    break
                print(error_message)

        plan_definition['name'] = app_name

        # Check for the language pack. Check args first, then planfile.
        # If it's neither of those places, prompt for it and update the
        # plan definition.

        languagepack = None
        if args.languagepack is not None:
            languagepack = args.languagepack
            plan_definition['artifacts'][0]['language_pack'] = languagepack
        elif plan_definition['artifacts'][0].get('language_pack') is None:
            languagepacks = self.client.languagepacks.list()
            filtered_list = cli_utils.filter_ready_lps(languagepacks)

            if len(filtered_list) > 0:
                lpnames = [lang_pack.name for lang_pack in filtered_list]
                lp_uuids = [lang_pack.uuid for lang_pack in filtered_list]
                fields = ['uuid', 'name', 'description',
                          'status', 'source_uri']
                self._print_list(filtered_list, fields)
                languagepack = input("Please choose a languagepack "
                                     "from the above list.\n> ")
                while languagepack not in lpnames + lp_uuids:
                    languagepack = input("You must choose one of "
                                         "the named language "
                                         "packs.\n> ")
                plan_definition['artifacts'][0]['language_pack'] = languagepack
            else:
                raise exc.CommandError("No languagepack in READY state. "
                                       "Create a languagepack first.")

        # Check for the git repo URL. Check args first, then the planfile.
        # If it's neither of those places, prompt for it and update the
        # plan definition.

        git_url = None
        if args.git_url is not None:
            plan_definition['artifacts'][0]['content']['href'] = args.git_url
        if plan_definition['artifacts'][0]['content'].get('href') is None:
            git_url = input("Please specify a git repository URL "
                            "for your application.\n> ")
            plan_definition['artifacts'][0]['content']['href'] = git_url
        git_url = plan_definition['artifacts'][0]['content']['href']

        # If a token is supplied, we won't need to generate one.
        artifact = plan_definition['artifacts'][0]
        repo_token = artifact.get('repo_token')

        is_private = (args.private_repo or
                      artifact['content'].get('private'))
        plan_definition['artifacts'][0]['content']['private'] = is_private

        git_url = transform_git_url(git_url, is_private)
        plan_definition['artifacts'][0]['content']['href'] = git_url

        # If we'll be adding a trigger, or the repo is private,
        # we'll need to use a personal access token.
        # The GitHubAuth object will create one if repo_token is null.
        if is_private or args.setup_trigger or args.workflow:
            gha = github.GitHubAuth(git_url, repo_token=repo_token)
            repo_token = repo_token or gha.repo_token
            # Created or provided, the repo token needs to be in the
            # plan data before we call client.plans.create.
            plan_definition['artifacts'][0]['repo_token'] = repo_token

        # Check for the entry point. Check args first, then the planfile.
        # If it's neither of those places, prompt for it and update the
        # plan definition.
        run_cmd = None
        if args.run_cmd is not None:
            plan_definition['artifacts'][0]['run_cmd'] = args.run_cmd
        if plan_definition['artifacts'][0].get('run_cmd') is None:
            run_cmd = input("Please specify start/run command for "
                            "your application.\n> ")
            plan_definition['artifacts'][0]['run_cmd'] = run_cmd

        # Check for unit test command
        if args.unittest_cmd is not None:
            plan_definition['artifacts'][0]['unittest_cmd'] = args.unittest_cmd

        # Check for the port.
        if args.port is not None:
            plan_definition['artifacts'][0]['ports'] = int(args.port)
        if plan_definition['artifacts'][0].get('ports') is None:
            plan_definition['artifacts'][0]['ports'] = 80

        # Update name and description if specified.
        a_name = ''
        if (plan_definition['artifacts'][0].get('name') is not None
                and plan_definition['artifacts'][0]['name'] != ''):
            a_name = plan_definition['artifacts'][0]['name']
            a_name = a_name.lower()
        else:
            a_name = app_name.lower()
        if not lpname_is_valid(a_name):
            # https://github.com/docker/compose/issues/941
            # Docker build only allows lowercase names for now.
            msg = name_error_message("Artifact names")
            raise exc.CommandError(msg)
        plan_definition['artifacts'][0]['name'] = a_name

        if args.desc is not None:
            plan_definition['description'] = args.desc
        elif plan_definition.get('description') is None:
            plan_definition['description'] = ''

        if args.param_file is not None:
            try:
                with open(args.param_file) as param_f:
                    param_def = param_f.read()
                plan_definition['parameters'] = yamlutils.load(param_def)
            except IOError:
                message = "Could not open param file %s." % args.param_file
                raise exc.CommandError(message=message)
            except ValueError:
                message = ("Param file %s was not YAML." %
                           args.param_file)
                raise exc.CommandError(message=message)

        plan = self.client.plans.create(yamlutils.dump(plan_definition))
        plan.status = 'REGISTERED'
        fields = ['uuid', 'name', 'description', 'uri', 'artifacts',
                  'trigger_uri', 'status']

        artifacts = copy.deepcopy(vars(plan).get('artifacts'))
        self._print_dict(plan, fields, wrap=72)

        # Solum generated a keypair; only upload the public key if we already
        # have a repo_token, since we'd only have one if we've authed against
        # github already.
        content = vars(artifacts[0]).get('content')
        if content:
            public_key = content.get('public_key', '')
            if repo_token and is_private and public_key:
                try:
                    gha = github.GitHubAuth(git_url, repo_token=repo_token)
                    gha.add_ssh_key(public_key=public_key)
                except github.GitHubException as ghe:
                    raise exc.CommandError(message=str(ghe))

        if args.setup_trigger or args.workflow:
            trigger_uri = vars(plan).get('trigger_uri', '')
            if trigger_uri:
                workflow = None
                if args.workflow:
                    workflow = args.workflow.replace('+', ' ').split(' ')
                try:
                    gha = github.GitHubAuth(git_url, repo_token=repo_token)
                    gha.create_webhook(trigger_uri, workflow=workflow)
                except github.GitHubException as ghe:
                    raise exc.CommandError(message=str(ghe))

    def deploy(self):
        """Deploy an application, building any applicable artifacts first."""
        # This is just "assembly create" with a little bit of introspection.
        # TODO(datsun180b): Add build() method, and add --build-id argument
        # to this method to allow for build-only and deploy-only workflows.
        self.parser.add_argument('app',
                                 help="Application name")
        self.parser._names['app'] = 'application'
        args = self.parser.parse_args()
        try:
            plan = self.client.plans.find(name_or_id=args.app)
        except exceptions.NotFound:
            message = "No app named '%s'." % args.app
            raise exceptions.NotFound(message=message)

        assembly = self.client.assemblies.create(name=plan.name,
                                                 description=plan.description,
                                                 plan_uri=plan.uri)
        fields = ['uuid', 'name', 'status', 'application_uri']
        self._print_dict(assembly, fields, wrap=72)

    def delete(self):
        """Delete an application and all related artifacts."""
        # This is "assembly delete" followed by "plan delete".
        self.parser.add_argument('app',
                                 help="Application name")
        self.parser._names['app'] = 'application'
        args = self.parser.parse_args()
        try:
            plan = self.client.plans.find(name_or_id=args.app)
        except exceptions.NotFound:
            message = "No app named '%s'." % args.app
            raise exceptions.NotFound(message=message)
        self.client.plans.delete(plan_id=str(plan.uuid))


class InfoCommands(cli_utils.NoSubCommands):
    """Show Solum server connection information.

Available commands:

    solum info
        Show Solum endpoint and API release version.

    """

    def info(self):
        parsed, _ = self.parser.parse_known_args()

        # Use the api_endpoint to get the catalog

        ks_kwargs = {
            'username': parsed.os_username,
            'password': parsed.os_password,
            'tenant_name': parsed.os_tenant_name,
            'auth_url': parsed.os_auth_url,
        }

        solum_api_endpoint = parsed.solum_url
        if not solum_api_endpoint:
            ksclient = keystoneclient.Client(**ks_kwargs)
            services = ksclient.auth_ref.service_catalog
            services = services.catalog['serviceCatalog']
            solum_service = [s for s in services if s['name'] == 'solum']
            try:
                endpoint = solum_service[0]['endpoints']
                solum_api_endpoint = endpoint[0]['publicURL']
            except (IndexError, KeyError):
                print("Error: SOLUM_URL not set, and no Solum endpoint "
                      "could be found in service catalog.")

        solum_api_version = ''
        if solum_api_endpoint:
            kwargs = {"disable_ssl_certificate_validation": not self.verify}
            try:
                resp, content = httplib2.Http(**kwargs).request(
                    solum_api_endpoint, 'GET')
                solum_api_version = resp.get('x-solum-release', '')
            except Exception:
                print("Error: Solum endpoint could not be contacted;"
                      " API version could not be determined.")

        print("python-solumclient version %s" % solumclient.__version__)
        print("solum API endpoint: %s" % solum_api_endpoint)
        print("solum API version: %s" % solum_api_version)
        print("solum auth endpoint: %s" % parsed.os_auth_url)
        print("solum auth username: %s" % parsed.os_username)
        print("solum auth tenant/project: %s" % parsed.os_tenant_name)


class PermissiveParser(argparse.ArgumentParser):
    """An ArgumentParser that handles errors without exiting.

    An argparse.ArgumentParser that doesn't sys.exit(2) when it
    gets the wrong number of arguments. Gives us better control
    over exception handling.

    """

    # Used in _check_positional_arguments to give a clearer name to missing
    # positional arguments.
    _names = None

    def __init__(self, *args, **kwargs):
        self._names = {}
        kwargs['add_help'] = False
        kwargs['description'] = argparse.SUPPRESS
        kwargs['usage'] = argparse.SUPPRESS
        super(PermissiveParser, self).__init__(*args, **kwargs)

    def error(self, message):
        raise exc.CommandError(message=message)

    def _report_missing_args(self):
        pass

    def parse_args(self, *args, **kwargs):
        ns, rem = self.parse_known_args(*args, **kwargs)
        if rem:
            unrec = ', '.join([a.split('=')[0].lstrip('-') for a in rem])
            raise exc.CommandError("Unrecognized arguments: %s" % unrec)
        return ns

    def parse_known_args(self, *args, **kwargs):
        ns, rem = argparse.Namespace(), []
        try:
            kwargs['namespace'] = ns
            ns, rem = super(PermissiveParser, self).parse_known_args(
                *args, **kwargs)
        except exc.CommandError:
            pass
        self._check_positional_arguments(ns)
        return ns, rem

    def _check_positional_arguments(self, namespace):
        for argument in self._positionals._group_actions:
            argname = argument.dest
            localname = self._names.get(argname, argname)
            article = 'an' if localname[0] in 'AEIOUaeiou' else 'a'
            if not vars(namespace).get(argname):
                message = 'You must specify %(article)s %(localname)s.'
                message %= {'article': article, 'localname': localname}
                raise exc.CommandError(message=message)


def main():
    """Solum command-line client.

Available commands:

    solum help
        Show this help message.

    solum info
        Show Solum endpoint and API release version.

    solum --version
        Show current Solum client version and exit.

    solum lp help
        Show a help message specific to languagepack commands.

    solum lp create <NAME> <GIT_REPO_URL>
        Create a new language pack from a git repo.

    solum lp list
        Print and index of all available language packs.

    solum lp show <NAME|UUID>
        Print the details of a language pack.

    solum lp delete <NAME|UUID>
        Destroy a language pack.

    solum lp logs <NAME|UUID>
        Show logs for a language pack.

    solum app help
        Show a help message specific to app commands.

    solum app list
        Print an index of all deployed applications.

    solum app show <NAME|UUID>
        Print detailed information about one application.

    solum app create [--app-file <AppFile>] [--git-url <GIT_URL>]
                     [--lp <LANGUAGEPACK>]
                     [--param-file <PARAMFILE>]
                     [--setup-trigger]
                     [--trigger-workflow <CUSTOM-WORKFLOW>]
                      <CUSTOM-WORKFLOW>=(unittest | build | unittest+build)
                      Without the --trigger-workflow flag,
                      the workflow unittest+build+deploy is triggered
                      (this is the default workflow)

        Register a new application with Solum.

    solum app deploy <NAME|UUID>
        Deploy an application, building any applicable artifacts first.
        du-id is optional flag. It can be used to pass in id of a previously
        created deployment unit. If passed, this command will deploy the du
        referenced by the provided du-id instead of building one first.

    solum app delete <NAME|UUID>
        Delete an application and all related artifacts.

    solum app logs <NAME|UUID> [--wf-id <wf-id>]
        Show the logs of an application for all the workflows.
        wf-id is optional flag which can be used to pass in id of one of
        the existing workflows. If provided, the logs only for that workflow
        are displayed.

    solum app scale <APP_NAME|UUID> <target>

    solum workflow list <APP_NAME|UUID>
        List all application workflows.

    solum workflow show <APP_NAME|UUID> <WORKFLOW_ID|UUID>
        Print the details of a workflow.

    solum workflow logs <APP_NAME|UUID> <WORKFLOW_ID|UUID>
        List all the logs of a given workflow.


    SOON TO BE DEPRECATED:

    solum oldapp create [--plan-file <PLANFILE>] [--git-url <GIT_URL>]
                     [--lp <LANGUAGEPACK>] [--run-cmd <RUN_CMD>]
                     [--unittest-cmd <UNITTEST_CMD>]
                     [--name <NAME>] [--port <PORT>]
                     [--param-file <PARAMFILE>]
                     [--desc <DESCRIPTION>]
                     [--setup-trigger]
                     [--private-repo]
                     [--trigger-workflow <WORKFLOW>]
        Register a new application with Solum.

    """

    parser = PermissiveParser()

    resources = {
        'oldapp': OldAppCommands,
        'app': AppCommands,
        'plan': PlanCommands,
        'pipeline': PipelineCommands,
        'lp': LanguagePackCommands,
        'languagepack': LanguagePackCommands,
        'component': ComponentCommands,
        'info': InfoCommands,
        'wf': WorkflowCommands,
        'workflow': WorkflowCommands,
    }

    choices = resources.keys()

    parser.add_argument('resource', choices=choices,
                        default='help',
                        help="Target noun to act upon")

    parser.add_argument('-V', '--version', action='store_true',
                        dest='show_version',
                        help="Report solum version.")

    parser.add_argument('-E', '--show-errors', dest='show_errors',
                        action='store_true',
                        help='Debug. Show traceback on error.')

    parsed, _ = parser.parse_known_args()

    if parsed.show_version:
        print("python-solumclient version %s" % solumclient.__version__)
        return

    resource = vars(parsed).get('resource')

    if resource in resources:
        if parsed.show_errors:
            resources[resource](parser)
        else:
            try:
                resources[resource](parser)
            except Exception as e:
                if hasattr(e, 'message'):
                    print("ERROR: %s" % e.message)
                else:
                    print("ERROR: %s" % e)

    else:
        print(main.__doc__)


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2013 - Noorul Islam K M
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from solumclient import client as solum_client
from solumclient.v1 import app
from solumclient.v1 import component
from solumclient.v1 import languagepack
from solumclient.v1 import pipeline
from solumclient.v1 import plan
from solumclient.v1 import platform
from solumclient.v1 import workflow


class Client(object):
    """Client for the Solum v1 API."""

    service_type = "application_deployment"

    def __init__(self, *args, **kwargs):
        """Initialize a new client for the Solum v1 API."""
        if not kwargs.get('auth_plugin'):
            kwargs['auth_plugin'] = solum_client.get_auth_plugin(**kwargs)
        self.auth_plugin = kwargs.get('auth_plugin')

        self.http_client = solum_client.construct_http_client(**kwargs)
        self.apps = app.AppManager(self.http_client)
        self.components = component.ComponentManager(self.http_client)
        self.pipelines = pipeline.PipelineManager(self.http_client)
        self.platform = platform.PlatformManager(self.http_client)
        self.plans = plan.PlanManager(self.http_client)
        self.languagepacks = languagepack.LanguagePackManager(self.http_client)
        self.workflows = workflow.WorkflowManager(self.http_client)



# Copyright 2014 - Rackspace Hosting
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from solumclient.common.apiclient import client
from solumclient.common.apiclient import exceptions
from solumclient.common.apiclient import fake_client
from solumclient.tests import base
from solumclient.v1 import pipeline

pipeline_list = [
    {
        'uri': 'http://example.com/v1/pipelines/x1',
        'name': 'database',
        'type': 'pipeline',
        'description': 'A mysql database',
        'tags': ['small'],
        'project_id': '1dae5a09ef2b4d8cbf3594b0eb4f6b94',
        'user_id': '55f41cf46df74320b9486a35f5d28a11',
        'component_links': [{
            'href': 'http://example.com:9777/v1/components/x1',
            'target_name': 'x1'}],
        'operations_uri': 'http://example.com:9777/v1/operations/o1',
        'sensors_uri': 'http://example.com:9777/v1/sensors/s1'
    },
    {
        'uri': 'http://example.com/v1/pipelines/x2',
        'name': 'load_balancer',
        'type': 'pipeline',
        'description': 'A load balancer',
        'tags': ['small'],
        'project_id': '1dae5a09ef2b4d8cbf3594b0eb4f6b94',
        'user_id': '55f41cf46df74320b9486a35f5d28a11',
        'component_links': [{
            'href': 'http://example.com:9777/v1/components/x2',
            'target_name': 'x2'}],
        'operations_uri': 'http://example.com:9777/v1/operations/o2',
        'sensors_uri': 'http://example.com:9777/v1/sensors/s2'
    }
]

pipeline_fixture = {
    'uri': 'http://example.com/v1/pipelines/x1',
    'name': 'database',
    'type': 'pipeline',
    'description': 'A mysql database',
    'tags': ['small'],
    'project_id': '1dae5a09ef2b4d8cbf3594b0eb4f6b94',
    'user_id': '55f41cf46df74320b9486a35f5d28a11',
    'component_links': [{
        'href': 'http://example.com:9777/v1/components/x1',
        'target_name': 'x1'}],
    'operations_uri': 'http://example.com:9777/v1/operations/o1',
    'sensors_uri': 'http://example.com:9777/v1/sensors/s1'
}

fixtures_list = {
    '/v1/pipelines': {
        'GET': (
            {},
            pipeline_list
        ),
    }
}


fixtures_get = {
    '/v1/pipelines/x1': {
        'GET': (
            {},
            pipeline_fixture
        ),
    }
}


fixtures_create = {
    '/v1/pipelines': {
        'POST': (
            {},
            pipeline_fixture
        ),
    }
}

fixtures_put = {
    '/v1/pipelines/x1': {
        'PUT': (
            {},
            pipeline_fixture
        ),
    }
}

fixtures_delete = {
    '/v1/pipelines/x1': {
        'DELETE': (
            {},
            {},
        ),
    }
}


class PipelineManagerTest(base.TestCase):

    def assert_pipeline_object(self, pipeline_obj):
        self.assertIn('Pipeline', repr(pipeline_obj))
        self.assertEqual(pipeline_fixture['uri'], pipeline_obj.uri)
        self.assertEqual(pipeline_fixture['type'], pipeline_obj.type)
        self.assertEqual(pipeline_fixture['project_id'],
                         pipeline_obj.project_id)
        self.assertEqual(pipeline_fixture['user_id'], pipeline_obj.user_id)

    def test_list_all(self):
        fake_http_client = fake_client.FakeHTTPClient(fixtures=fixtures_list)
        api_client = client.BaseClient(fake_http_client)
        mgr = pipeline.PipelineManager(api_client)
        pipelines = mgr.list()
        self.assertEqual(2, len(pipelines))
        self.assertIn('Pipeline', repr(pipelines[0]))
        self.assertEqual(pipeline_list[0]['uri'], pipelines[0].uri)
        self.assertEqual(pipeline_list[1]['uri'], pipelines[1].uri)

    def test_find_one(self):
        fake_http_client = fake_client.FakeHTTPClient(fixtures=fixtures_list)
        api_client = client.BaseClient(fake_http_client)
        mgr = pipeline.PipelineManager(api_client)
        pipelines = mgr.findall(name='database')
        self.assertEqual(1, len(pipelines))
        self.assertIn('Pipeline', repr(pipelines[0]))
        self.assertEqual(pipeline_list[0]['uri'], pipelines[0].uri)

    def test_find_one_only(self):
        fake_http_client = fake_client.FakeHTTPClient(fixtures=fixtures_list)
        api_client = client.BaseClient(fake_http_client)
        mgr = pipeline.PipelineManager(api_client)
        result = mgr.find(name_or_id='database')
        self.assertEqual(pipeline_list[0]['uri'], result.uri)

    def test_find_none(self):
        fake_http_client = fake_client.FakeHTTPClient(fixtures=fixtures_list)
        api_client = client.BaseClient(fake_http_client)
        mgr = pipeline.PipelineManager(api_client)
        self.assertRaises(exceptions.NotFound, mgr.find, name_or_id='what')

    def test_create(self):
        fake_http_client = fake_client.FakeHTTPClient(fixtures=fixtures_create)
        api_client = client.BaseClient(fake_http_client)
        mgr = pipeline.PipelineManager(api_client)
        pipeline_obj = mgr.create()
        self.assert_pipeline_object(pipeline_obj)

    def test_get(self):
        fake_http_client = fake_client.FakeHTTPClient(fixtures=fixtures_get)
        api_client = client.BaseClient(fake_http_client)
        mgr = pipeline.PipelineManager(api_client)
        pipeline_obj = mgr.get(pipeline_id='x1')
        self.assert_pipeline_object(pipeline_obj)

    def test_put(self):
        fake_http_client = fake_client.FakeHTTPClient(fixtures=fixtures_put)
        api_client = client.BaseClient(fake_http_client)
        mgr = pipeline.PipelineManager(api_client)
        pipeline_obj = mgr.put(pipeline_id='x1')
        self.assert_pipeline_object(pipeline_obj)

    def test_delete(self):
        fake_http_client = fake_client.FakeHTTPClient(fixtures=fixtures_delete)
        api_client = client.BaseClient(fake_http_client)
        mgr = pipeline.PipelineManager(api_client)
        mgr.delete(pipeline_id='x1')
        fake_http_client.assert_called('DELETE', '/v1/pipelines/x1')



# Copyright 2013 - Noorul Islam K M
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from io import StringIO
from unittest import mock

import collections
import re
import sys

import fixtures

from oslo_utils import uuidutils

from stevedore import extension
import testtools
from testtools import matchers

from solumclient import client
from solumclient.common.apiclient import auth
from solumclient.common import yamlutils
from solumclient import solum
from solumclient.tests import base
from solumclient.v1 import component
from solumclient.v1 import languagepack
from solumclient.v1 import pipeline
from solumclient.v1 import plan

FAKE_ENV = {'OS_USERNAME': 'username',
            'OS_PASSWORD': 'password',
            'OS_TENANT_NAME': 'tenant_name',
            'OS_AUTH_URL': 'http://no.where'}

languagepack_file_data = (
    '{"language-pack-type":"Java", "language-pack-name":"Java version 1.4."}')


class MockEntrypoint(object):
    def __init__(self, name, plugin):
        self.name = name
        self.plugin = plugin


class BaseFakePlugin(auth.BaseAuthPlugin):
    def _do_authenticate(self, http_client):
        pass

    def token_and_endpoint(self, endpoint_type, service_type):
        pass


class TestSolum(base.TestCase):
    """Test the Solum CLI."""

    re_options = re.DOTALL | re.MULTILINE

    # Patch os.environ to avoid required auth info.
    def make_env(self, exclude=None):
        env = dict((k, v) for k, v in FAKE_ENV.items() if k != exclude)
        self.useFixture(fixtures.MonkeyPatch('os.environ', env))

    @mock.patch.object(extension.ExtensionManager, "map")
    def shell(self, argstr, mock_mgr_map, exit_code=0):
        class FakePlugin(BaseFakePlugin):
            def authenticate(self, cls):
                cls.request(
                    "POST", "http://auth/tokens",
                    json={"fake": "me"}, allow_redirects=True)

        mock_mgr_map.side_effect = (
            lambda func: func(MockEntrypoint("fake", FakePlugin)))

        orig = sys.stdout
        try:
            sys.stdout = StringIO()
            argv = [__file__, ]
            argv.extend(argstr.split())
            self.useFixture(
                fixtures.MonkeyPatch('sys.argv', argv))

            solum.main()
        except SystemExit:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            self.assertEqual(exit_code, exc_value.code)
        finally:
            out = sys.stdout.getvalue()
            sys.stdout.close()
            sys.stdout = orig

        return out

    @testtools.skip("Skip it when found solution")
    def test_get_client_debug(self):
        env = dict((k, v) for k, v in FAKE_ENV.items() if k is not None)
        test_client = client.get_client('1', **env)
        self.assertFalse(test_client.http_client.debug)
        test_client = client.get_client('1', debug=True)
        self.assertTrue(test_client.http_client.debug)

    @testtools.skip("Skip it when found solution")
    def test_get_client_insecure(self):
        env = dict((k, v) for k, v in FAKE_ENV.items() if k is not None)
        test_client = client.get_client('1', **env)
        self.assertTrue(test_client.http_client.verify)
        test_client = client.get_client('1', verify=False)
        self.assertFalse(test_client.http_client.verify)

    def test_help(self):
        required = [
            '.*?^Available commands:'

        ]
        help_text = self.shell('--help')
        for r in required:
            self.assertThat(help_text,
                            matchers.MatchesRegex(r,
                                                  self.re_options))

    # Workflow Tests #
    def test_workflow_error(self):
        self.make_env()
        out = self.shell("workflow unknown-command")
        self.assertIn("Available commands", out)

    # Pipeline Tests #
    @mock.patch.object(pipeline.PipelineManager, "list")
    def test_pipeline_list(self, mock_pipeline_list):
        self.make_env()
        self.shell("pipeline list")
        mock_pipeline_list.assert_called_once_with()

    @mock.patch.object(pipeline.PipelineManager, "create")
    def test_pipeline_create(self, mock_pipeline_create):
        self.make_env()
        self.shell("pipeline create http://example.com/a.yaml workbook test")
        mock_pipeline_create.assert_called_once_with(
            name='test',
            workbook_name='workbook',
            plan_uri='http://example.com/a.yaml')

    @mock.patch.object(pipeline.PipelineManager, "create")
    def test_pipeline_create_without_name(self, mock_pipeline_create):
        self.make_env()
        self.shell("pipeline create http://example.com/a.yaml workbook",
                   exit_code=2)

    @mock.patch.object(plan.PlanManager, "find")
    @mock.patch.object(pipeline.PipelineManager, "create")
    def test_pipeline_create_with_plan_name(self, mock_pipeline_create,
                                            mock_plan_find):
        class FakePlan(object):
            uri = 'http://example.com/the-plan.yaml'

        self.make_env()
        mock_plan_find.return_value = FakePlan()
        self.shell("pipeline create the-plan-name workbook test")
        mock_plan_find.assert_called_once_with(name_or_id='the-plan-name')
        mock_pipeline_create.assert_called_once_with(
            name='test',
            workbook_name='workbook',
            plan_uri='http://example.com/the-plan.yaml')

    @mock.patch.object(pipeline.PipelineManager, "delete")
    @mock.patch.object(pipeline.PipelineManager, "find")
    def test_pipeline_delete(self, mock_pipeline_find, mock_pipeline_delete):
        self.make_env()
        the_id = uuidutils.generate_uuid()
        self.shell("pipeline delete %s" % the_id)
        mock_pipeline_find.assert_called_once_with(
            name_or_id=the_id)
        self.assertEqual(1, mock_pipeline_delete.call_count)

    @mock.patch.object(pipeline.PipelineManager, "find")
    def test_pipeline_get(self, mock_pipeline_find):
        self.make_env()
        the_id = uuidutils.generate_uuid()
        self.shell("pipeline show %s" % the_id)
        mock_pipeline_find.assert_called_once_with(name_or_id=the_id)

    @mock.patch.object(pipeline.PipelineManager, "find")
    def test_pipeline_get_by_name(self, mock_pipeline_find):
        self.make_env()
        self.shell("pipeline show app2")
        mock_pipeline_find.assert_called_once_with(name_or_id='app2')

    # App Tests #
    def test_app_create_with_missing_workflow_config(self):
        raw_data = 'version: 1\nname: ex_app\nworkflow_config:\n'
        mopen = mock.mock_open(read_data=raw_data)

        with mock.patch('%s.open' % solum.__name__, mopen, create=True):
            self.make_env()
            out = self.shell("app create --app-file /dev/null")
            self.assertEqual("ERROR: Workflow config cannot be empty\n", out)

    def test_app_create_with_missing_trigger_actions(self):
        raw_data = 'version: 1\nname: ex_app\ntrigger_actions:\n'
        mopen = mock.mock_open(read_data=raw_data)

        with mock.patch('%s.open' % solum.__name__, mopen, create=True):
            self.make_env()
            out = self.shell("app create --app-file /dev/null")
            self.assertEqual("ERROR: Trigger actions cannot be empty\n", out)

    def test_app_create_with_bad_name(self):
        raw_data = '\n'.join([
            'version: 1',
            'name: ex=app1',
            'description: python web app',
            'workflow_config:',
            '  run_cmd: python app.py',
            'ports: [5000]'])

        mopen = mock.mock_open(read_data=raw_data)

        with mock.patch('%s.open' % solum.__name__, mopen, create=True):
            self.make_env()
            out = self.shell("app create --app-file /dev/null")

            self.assertEqual("ERROR: Application name must be 1-100 "
                             "characters long, only contain a-z,0-9,-,_ and "
                             "start with an alphabet character.\n", out)

    # OldApp Tests #
    def test_oldapp_create_with_missing_artifacts(self):
        raw_data = 'version: 1\nname: ex_plan1\ndescription: dsc1.'
        mopen = mock.mock_open(read_data=raw_data)

        with mock.patch('%s.open' % solum.__name__, mopen, create=True):
            self.make_env()
            out = self.shell("oldapp create --plan-file /dev/null")
            self.assertEqual("ERROR: Missing artifacts section\n", out)

    def test_oldapp_create_with_bad_name(self):
        raw_data = '\n'.join([
            'version: 1',
            'name: ex=plan1',
            'description: python web app',
            'artifacts:',
            '- name: web',
            '  content:',
            '    href: https://github.com/user/repo.git',
            '  language_pack: auto',
            '  unittest_cmd: ./unit_tests.sh',
            '  run_cmd: python app.py',
            '  ports: 5000'])

        mopen = mock.mock_open(read_data=raw_data)

        with mock.patch('%s.open' % solum.__name__, mopen, create=True):
            self.make_env()
            out = self.shell("oldapp create --plan-file /dev/null")

            self.assertEqual("ERROR: Application name must be 1-100 "
                             "characters long, only contain a-z,0-9,-,_ and "
                             "start with an alphabet character.\n", out)

    @testtools.skip("Skip it when found solution")
    def test_oldapp_create_with_bad_artifact_name(self):
        raw_data = '\n'.join([
            'version: 1',
            'name: explan1',
            'description: python web app',
            'artifacts:',
            '- name:',
            '  content:',
            '    href: https://github.com/user/repo.git',
            '  language_pack: auto',
            '  unittest_cmd: ./unit_tests.sh',
            '  run_cmd: python app.py',
            '  ports: 5000'])

        mopen = mock.mock_open(read_data=raw_data)

        with mock.patch('%s.open' % solum.__name__, mopen, create=True):
            self.make_env()
            out = self.shell("oldapp create --plan-file /dev/null")

            # No part of the plan is in error; the next step in the test
            # is authorization, which is deliberately mocked.
            self.assertIn("ERROR: Authorization Failed:", out)
            self.assertIn("http://no.where/tokens", out)

    def test_oldapp_create_with_artifacts_empty(self):
        raw_data = 'version: 1\nname: ex_plan1\ndescription: dsc1.\nartifacts:'
        mopen = mock.mock_open(read_data=raw_data)

        with mock.patch('%s.open' % solum.__name__, mopen, create=True):
            self.make_env()
            out = self.shell("oldapp create --plan-file /dev/null")
            self.assertEqual("ERROR: Artifacts cannot be empty\n", out)

    def test_oldapp_create_with_artifacts_no_content(self):
        raw_data = 'version: 1\nname: ex_plan1\ndescription: d1.\nartifacts:\n'
        raw_data += '- name: asdfds\n'
        raw_data += '  artifact_type: heroku\n'
        raw_data += '  language_pack: lp'

        mopen = mock.mock_open(read_data=raw_data)

        with mock.patch('%s.open' % solum.__name__, mopen, create=True):
            self.make_env()
            out = self.shell("oldapp create --plan-file /dev/null")
            self.assertEqual("ERROR: Artifact content missing\n", out)

    # Plan Tests #
    @mock.patch.object(plan.PlanManager, "create")
    def test_plan_create(self, mock_plan_create):
        FakeResource = collections.namedtuple("FakeResource",
                                              "uuid name description uri")

        mock_plan_create.return_value = FakeResource('foo', 'foo', 'foo',
                                                     'foo')
        raw_data = 'version: 1\nname: ex_plan1\ndescription: dsc1.'
        plan_data = yamlutils.dump(yamlutils.load(raw_data))
        mopen = mock.mock_open(read_data=raw_data)
        with mock.patch('%s.open' % solum.__name__, mopen, create=True):
            self.make_env()
            self.shell("plan create /dev/null")
            mock_plan_create.assert_called_once_with(plan_data)

    @mock.patch.object(plan.PlanManager, "create")
    def test_plan_create_with_private_github_repo(self, mock_plan_create):
        FakeResource = collections.namedtuple(
            "FakeResource", "uuid name description uri artifacts")

        mock_plan_create.return_value = FakeResource('foo', 'foo', 'foo',
                                                     'foo', 'artifacts')
        expected_printed_dict_args = mock_plan_create.return_value._asdict()
        expected_printed_dict_args.pop('artifacts')
        raw_data = 'version: 1\nname: ex_plan1\ndescription: dsc1.'
        plan_data = yamlutils.dump(yamlutils.load(raw_data))
        mopen = mock.mock_open(read_data=raw_data)
        with mock.patch('%s.open' % solum.__name__, mopen, create=True):
            self.make_env()
            self.shell("plan create /dev/null")
            mock_plan_create.assert_called_once_with(plan_data)

    @mock.patch.object(plan.PlanManager, "list")
    def test_plan_list(self, mock_plan_list):
        self.make_env()
        self.shell("plan list")
        mock_plan_list.assert_called_once_with()

    @mock.patch.object(plan.PlanManager, "delete")
    @mock.patch.object(plan.PlanManager, "find")
    def test_plan_delete(self, mock_plan_find, mock_plan_delete):
        self.make_env()
        the_id = uuidutils.generate_uuid()
        self.shell("plan delete %s" % the_id)
        mock_plan_find.assert_called_once_with(name_or_id=the_id)
        self.assertEqual(1, mock_plan_delete.call_count)

    @mock.patch.object(plan.PlanManager, "find")
    def test_plan_get(self, mock_plan_find):
        self.make_env()
        the_id = uuidutils.generate_uuid()
        self.shell("plan show %s" % the_id)
        mock_plan_find.assert_called_once_with(name_or_id=the_id)

    @mock.patch.object(plan.PlanManager, "find")
    def test_plan_get_private_github_repo(self, mock_plan_find):
        self.make_env()
        the_id = uuidutils.generate_uuid()
        FakeResource = collections.namedtuple(
            "FakeResource", "uuid name description uri artifacts")
        mock_plan_find.return_value = FakeResource('foo', 'foo', 'foo', 'foo',
                                                   'artifacts')
        self.shell("plan show %s" % the_id)
        mock_plan_find.assert_called_once_with(name_or_id=the_id)

    # LanguagePack Tests #
    @mock.patch.object(languagepack.LanguagePackManager, "list")
    def test_languagepack_list(self, mock_lp_list):
        self.make_env()
        self.shell("languagepack list")
        self.assertEqual(1, mock_lp_list.call_count)

    @mock.patch.object(languagepack.LanguagePackManager, "create")
    def test_languagepack_create(self, mock_lp_create):
        FakeResource = collections.namedtuple("FakeResource",
                                              "uuid name description "
                                              "compiler_versions os_platform")
        mock_lp_create.return_value = FakeResource(
            'foo', 'foo', 'foo', 'foo', 'foo')
        lp_metadata = '{"OS": "Ubuntu"}'
        mopen = mock.mock_open(read_data=lp_metadata)
        with mock.patch('%s.open' % solum.__name__, mopen, create=True):
            self.make_env()

            self.shell("languagepack create lp_name github.com/test "
                       "--lp_metadata=/dev/null")
            mock_lp_create.assert_called_once_with(
                name='lp_name',
                source_uri='github.com/test',
                lp_metadata=lp_metadata,
                lp_params={})

    @mock.patch.object(languagepack.LanguagePackManager, "delete")
    def test_languagepack_delete(self, mock_lp_delete):
        self.make_env()
        self.shell("languagepack delete fake-lp-id")
        mock_lp_delete.assert_called_once_with(lp_id='fake-lp-id')

    @mock.patch.object(languagepack.LanguagePackManager, "find")
    def test_languagepack_get(self, mock_lp_get):
        self.make_env()
        self.shell("languagepack show fake-lp-id1")
        mock_lp_get.assert_called_once_with(name_or_id='fake-lp-id1')

    # Component Tests #
    @mock.patch.object(component.ComponentManager, "list")
    def test_component_list(self, mock_component_list):
        self.make_env()
        self.shell("component list")
        mock_component_list.assert_called_once_with()

    @mock.patch.object(component.ComponentManager, "find")
    def test_component_get(self, mock_component_find):
        self.make_env()
        the_id = uuidutils.generate_uuid()
        self.shell("component show %s" % the_id)
        mock_component_find.assert_called_once_with(name_or_id=the_id)

    @mock.patch.object(component.ComponentManager, "find")
    def test_component_get_by_name(self, mock_component_find):
        self.make_env()
        self.shell("component show comp1")
        mock_component_find.assert_called_once_with(name_or_id='comp1')

    def test_transform_git_url(self):
        private_uri = 'git@github.com:solum/python.git'
        public_uri = 'https://github.com/solum/python.git'
        private = True
        public = False

        result = solum.transform_git_url(private_uri, private)
        self.assertEqual(result, private_uri)

        result = solum.transform_git_url(public_uri, private)
        self.assertEqual(result, private_uri)

        result = solum.transform_git_url(private_uri, public)
        self.assertEqual(result, public_uri)

        result = solum.transform_git_url(public_uri, public)
        self.assertEqual(result, public_uri)



#!/usr/bin/env python
# Copyright 2014 - Rackspace Hosting
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Setup everything for a Github-hosted repo to use solum as its CI tool.
"""

import argparse
import base64
import getpass
import json
import os
import random
import re
import string
import tempfile

import httplib2
import six
import yaml

from solumclient import client as solum_client
from solumclient.common import cliutils


SOLUM_API_VERSION = '1'
CREDENTIALS = {}
PLAN_TEMPLATE = {"version": 1,
                 "name": "chef",
                 "description": "chef testy",
                 "artifacts": []}


def _get_solum_client():
    args = {}
    args['os_username'] = os.getenv('OS_USERNAME', '')
    args['os_password'] = os.getenv('OS_PASSWORD', '')
    args['os_tenant_name'] = os.getenv('OS_TENANT_NAME', '')
    args['os_auth_url'] = os.getenv('OS_AUTH_URL', '')
    args['solum_url'] = os.getenv('SOLUM_URL', '')

    try:
        client = solum_client.get_client(SOLUM_API_VERSION, **args)
        return client
    except Exception as ex:
        print("Error in getting Solum client: %s" % ex)
        exit(1)


def _get_token(git_url):
    # Get an OAuth token with the scope of 'repo' for the user
    if git_url in CREDENTIALS and 'token' in CREDENTIALS[git_url]:
        return CREDENTIALS[git_url]['token']

    repo_pat = re.compile(r'github\.com[:/](.+?)/(.+?)($|/$|\.git$|\.git/$)')
    match = repo_pat.search(git_url)
    if match:
        user_org_name = match.group(1)
        repo = match.group(2)
    else:
        print('Failed parsing %s' % git_url)
        exit(1)

    full_repo_name = '/'.join([user_org_name, repo])
    username = six.moves.input("Username for repo '%s' [%s]: " %
                               (full_repo_name, user_org_name))
    if not username:
        username = user_org_name
    password = getpass.getpass("Password: ")
    # TODO(james_li): add support for two-factor auth
    CREDENTIALS[git_url] = {}
    CREDENTIALS[git_url]['user'] = username
    CREDENTIALS[git_url]['password'] = password
    CREDENTIALS[git_url]['full_repo'] = full_repo_name

    http = httplib2.Http()
    auth = base64.encodestring(username + ':' + password)
    headers = {'Authorization': 'Basic ' + auth,
               'Content-Type': 'application/json'}
    # 'note' field has to be unique
    note = 'Solum-status-' + ''.join(random.sample(string.lowercase, 5))
    data = {'scopes': 'repo', 'note': note}

    # TODO(james_li): make the url configurable
    resp, content = http.request('https://api.github.com/authorizations',
                                 'POST', headers=headers,
                                 body=json.dumps(data))

    if resp['status'] == '201' or resp['status'] == '200':
        content_dict = json.loads(content)
        CREDENTIALS[git_url]['token'] = str(content_dict['token'])
        return CREDENTIALS[git_url]['token']
    else:
        print('Failed to get token from Github')
        exit(1)


def _filter_trigger_url(url):
    filtered_url = url
    url_pattern = re.compile(r'^(http://)(.+)')
    match = url_pattern.search(url)
    if match:
        filtered_url = ''.join(['https://', match.group(2)])
    else:
        print('Cannot filter trigger url, to use the original one')
    return filtered_url


def get_planfile(git_uri, app_name, cmd, public):
    plan_dict = dict.copy(PLAN_TEMPLATE)
    plan_dict['name'] = app_name
    plan_dict['description'] = git_uri  # Put repo uri as plan desc.
    arti = {"name": "chef", "artifact_type": "chef",
            "content": {}, "language_pack": "auto"}
    arti['content']['href'] = git_uri
    if not public:
        arti['content']['private'] = True
    arti['unittest_cmd'] = cmd
    plan_dict['artifacts'].append(arti)

    # Create a Github token and insert it into plan file
    for arti in plan_dict['artifacts']:
        arti['status_token'] = _get_token(arti['content']['href'])
    plan_file = tempfile.NamedTemporaryFile(suffix='.yaml',
                                            prefix='solum_',
                                            delete=False)
    plan_file.write(yaml.dump(plan_dict, default_flow_style=False))
    plan_file_name = plan_file.name
    plan_file.close()
    return plan_file_name


def create_plan(client, plan_file):
    cmd = ['solum', 'app', 'create', plan_file]
    print(' '.join(cmd))
    with open(plan_file) as definition_file:
        definition = definition_file.read()

    plan = client.plans.create(definition)
    fields = ['uuid', 'name', 'description', 'uri']
    data = dict([(f, getattr(plan, f, ''))
                 for f in fields])
    cliutils.print_dict(data, wrap=72)

    if data['uri'] is None:
        print('Error: no uri found in plan creation')
        exit(1)

    # get public keys in the case of private repos
    artifacts = getattr(plan, 'artifacts', [])
    for arti in artifacts:
        content = getattr(arti, 'content', {})
        if 'public_key' in content and 'href' in content:
            CREDENTIALS[content['href']]['pub_key'] = content['public_key']

    return data['uri']


def create_assembly(client, app_name, plan_uri):
    cmd = ['solum', 'assembly', 'create', app_name, plan_uri]
    print(' '.join(cmd))
    assembly = client.assemblies.create(name=app_name, plan_uri=plan_uri)

    fields = ['uuid', 'name', 'description', 'status', 'application_uri',
              'trigger_uri']
    data = dict([(f, getattr(assembly, f, ''))
                 for f in fields])
    cliutils.print_dict(data, wrap=72)

    trigger_uri = data['trigger_uri']
    if trigger_uri is None:
        print('Error in trigger uri')
        exit(1)

    return trigger_uri


def create_webhook(trigger_uri):
    # Create github web hooks for pull requests
    for key in CREDENTIALS.keys():
        user = CREDENTIALS[key]['user']
        password = CREDENTIALS[key]['password']
        auth = base64.encodestring(user + ':' + password)
        http = httplib2.Http()
        # TODO(james_li): make this url configurable
        github_url = ('https://api.github.com/repos/%s/hooks' %
                      CREDENTIALS[key]['full_repo'])
        headers = {'Authorization': 'Basic ' + auth,
                   'Content-Type': 'application/json'}
        data = {'name': 'web',
                'events': ['pull_request', 'commit_comment'],
                'config': {'content_type': 'json',
                           'url': trigger_uri}}

        resp, _ = http.request(github_url, 'POST',
                               headers=headers,
                               body=json.dumps(data))

        if resp['status'] != '201' and resp['status'] != '200':
            print("Failed to create web hooks")
            print("Make sure you have access to repo '%s'" % key)
            exit(1)


def add_ssh_keys(args):
    if args.public:
        return

    # add public keys
    for key in CREDENTIALS.keys():
        if ('pub_key' in CREDENTIALS[key] and
                CREDENTIALS[key]['pub_key'] is not None):
            user = CREDENTIALS[key]['user']
            password = CREDENTIALS[key]['password']
            auth = base64.encodestring(user + ':' + password)
            http = httplib2.Http()
            if args.user_key:
                # TODO(james_li): make the url configurable
                github_url = 'https://api.github.com/user/keys'
            else:
                github_url = ('https://api.github.com/repos/%s/keys' %
                              CREDENTIALS[key]['full_repo'])
            headers = {'Authorization': 'Basic ' + auth,
                       'Content-Type': 'application/json'}
            data = {'title': 'devops@Solum',
                    'key': CREDENTIALS[key]['pub_key']}

            resp, _ = http.request(github_url, 'POST',
                                   headers=headers,
                                   body=json.dumps(data))

            if resp['status'] != '201' and resp['status'] != '200':
                if args.user_key:
                    print("Failed to add a ssh key to the account %s" % user)
                else:
                    print("Failed to add a deploy key to the repo %s" % key)
                exit(1)


def validate_args(args):
    if len(args.command) == 0 or len(args.git_uri) == 0:
        print("Please input for --test-cmd and --git-uri")
        exit(1)

    # try to use a correct git uri
    pat = re.compile(r'github\.com[:/](.+?)/(.+?)($|/.*$|\.git$|\.git/.*$)')
    match = pat.search(args.git_uri)
    if match:
        user_org_name = match.group(1)
        repo = match.group(2)
        if args.public:
            correct_uri = 'https://github.com/%s/%s' % (user_org_name, repo)
        else:
            correct_uri = 'git@github.com:%s/%s.git' % (user_org_name, repo)
        return correct_uri
    else:
        print("The input git uri seems not right")
        if args.public:
            print("The correct format is: https://github.com/<USER>/<REPO>")
        else:
            print("The correct format is: git@github.com:<USER>/<REPO>.git")
        exit(1)


def main(args):
    git_uri = validate_args(args)
    client = _get_solum_client()
    plan_file = get_planfile(git_uri, args.app_name, args.command, args.public)
    print('\n')
    print("************************* Starting setup *************************")
    print('\n')
    plan_uri = create_plan(client, plan_file)
    add_ssh_keys(args)
    try:
        os.remove(plan_file)
    except OSError:
        print('Cannot remove %s. Skip and move forward...' % plan_file)

    trigger_uri = create_assembly(client, args.app_name, plan_uri)
    create_webhook(_filter_trigger_url(trigger_uri))
    print('Successfully created Solum plan, assembly and webhooks!')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('app_name', help="app name")
    parser.add_argument('--git-uri', required=True, dest='git_uri',
                        help="git repo uri")
    parser.add_argument('--test-cmd', required=True, dest='command',
                        help="entrypoint to run tests")
    parser.add_argument('--public', action='store_true', default=False,
                        dest='public', help="public repo, defaults to False")
    parser.add_argument('--user-key', action='store_true', default=False,
                        dest='user_key', help="add SSH key to the user account"
                                              " rather than the repo,"
                                              " defaults to False")

    args = parser.parse_args()
    main(args)



# Copyright 2013 - Noorul Islam K M
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from solumclient.common.apiclient import exceptions


class CommandException(Exception):
    """An error occurred."""
    def __init__(self, message=None):
        self.message = message

    def __str__(self):
        return self.message or self.__class__.__doc__


class CommandError(CommandException):
    """Invalid usage of CLI."""


class NotUnique(CommandException):
    """Name refers to more than one of a given resource."""
    def __init__(self, resource='resource'):
        message = "More than one %s by that name. Retry with the UUID."
        self.message = message % resource


def from_response(response, method, url):
    """Returns an instance of :class:`HttpError` or subclass based on response.

    :param response: instance of `requests.Response` class
    :param method: HTTP method used for request
    :param url: URL used for request
    """
    kwargs = {
        "http_status": response.status_code,
        "response": response,
        "method": method,
        "url": url,
        "request_id": response.headers.get("x-compute-request-id"),
    }
    if "retry-after" in response.headers:
        kwargs["retry_after"] = response.headers["retry-after"]

    content_type = response.headers.get("Content-Type", "")
    if content_type.startswith("application/json"):
        try:
            body = response.json()
        except ValueError:
            pass
        else:
            if isinstance(body, dict):
                kwargs["message"] = body.get("faultstring")
                kwargs["details"] = body.get("debuginfo")
    elif content_type.startswith("text/"):
        kwargs["details"] = response.text

    try:
        cls = exceptions._code_map[response.status_code]
    except KeyError:
        if 500 <= response.status_code < 600:
            cls = exceptions.HttpServerError
        elif 400 <= response.status_code < 500:
            cls = exceptions.HTTPClientError
        else:
            cls = exceptions.HttpError
    return cls(**kwargs)



# Copyright 2014 - Noorul Islam K M
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from solumclient.builder import client
from solumclient.common.apiclient import exceptions
from solumclient.common import auth
from solumclient.tests import base


class ClientTest(base.TestCase):

    def test_client_unsupported_version(self):
        self.assertRaises(exceptions.UnsupportedVersion,
                          client.Client, '111.11', **{})

    def test_client(self):
        with mock.patch.object(auth, 'KeystoneAuthPlugin'):
            client.Client('1', **{})



# Copyright 2013 - Noorul Islam K M
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import time

from oslo_log import log as logging

from solumclient.common.apiclient import client as api_client
from solumclient.common import exc
from solumclient import config


_logger = logging.getLogger(__name__)


class HTTPClient(api_client.HTTPClient):
    def request(self, method, url, **kwargs):
        """Send an http request with the specified characteristics.

        Wrapper around `requests.Session.request` to handle tasks such as
        setting headers, JSON encoding/decoding, and error handling.

        :param method: method of HTTP request
        :param url: URL of HTTP request
        :param kwargs: any other parameter that can be passed to
'            requests.Session.request (such as `headers`) or `json`
             that will be encoded as JSON and used as `data` argument
        """

        kwargs.setdefault("headers", kwargs.get("headers", {}))
        kwargs["headers"]["User-Agent"] = self.user_agent
        kwargs["headers"]["X-User-ID"] = config.username
        kwargs["headers"]["X-Password"] = config.password
        kwargs["headers"]["X-Project"] = config.tenant
        if self.original_ip:
            kwargs["headers"]["Forwarded"] = "for=%s;by=%s" % (
                self.original_ip, self.user_agent)
        if self.timeout is not None:
            kwargs.setdefault("timeout", self.timeout)
        kwargs.setdefault("verify", self.verify)
        if self.cert is not None:
            kwargs.setdefault("cert", self.cert)
        self.serialize(kwargs)

        self._http_log_req(method, url, kwargs)
        if self.timings:
            start_time = time.time()
        resp = self.http.request(method, url, **kwargs)
        if self.timings:
            self.times.append(("%s %s" % (method, url),
                               start_time, time.time()))
        self._http_log_resp(resp)

        if resp.status_code >= 400:
            _logger.debug(
                "Request returned failure status: %s",
                resp.status_code)
            raise exc.from_response(resp, method, url)

        return resp



# Copyright 2013 - Noorul Islam K M
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import requests
from unittest import mock

from solumclient.common.apiclient import auth
from solumclient.common.apiclient import client as api_client
from solumclient.common.apiclient import exceptions
from solumclient.common import client
from solumclient.tests import base


class TestClient(api_client.BaseClient):
    service_type = "test"


class FakeAuthPlugin(auth.BaseAuthPlugin):
    auth_system = "fake"
    attempt = -1

    def _do_authenticate(self, http_client):
        pass

    def token_and_endpoint(self, endpoint_type, service_type):
        self.attempt = self.attempt + 1
        return ("token-%s" % self.attempt, "/endpoint-%s" % self.attempt)


class ClientTest(base.TestCase):
    def test_client_request(self):
        http_client = client.HTTPClient(FakeAuthPlugin())
        mock_request = mock.Mock()
        mock_request.return_value = requests.Response()
        mock_request.return_value.status_code = 200
        with mock.patch("requests.Session.request", mock_request):
            http_client.client_request(
                TestClient(http_client), "GET", "/resource", json={"1": "2"})
            requests.Session.request.assert_called_with(
                "GET",
                "/endpoint-0/resource",
                headers={
                    "User-Agent": http_client.user_agent,
                    "Content-Type": "application/json",
                    "X-Auth-Token": "token-0",
                    "X-Password": "",
                    "X-User-ID": "",
                    "X-Project": ""
                },
                data='{"1": "2"}',
                verify=True)

    def test_client_with_response_404_status_code(self):
        http_client = client.HTTPClient(FakeAuthPlugin())
        mock_request = mock.Mock()
        mock_request.return_value = requests.Response()
        mock_request.return_value.status_code = 404
        with mock.patch("requests.Session.request", mock_request):
            self.assertRaises(
                exceptions.HttpError, http_client.client_request,
                TestClient(http_client), "GET", "/resource")

    def test_client_with_invalid_endpoint(self):
        http_client = client.HTTPClient(FakeAuthPlugin())
        mock_request = mock.Mock()
        mock_request.side_effect = requests.ConnectionError
        with mock.patch("requests.Session.request", mock_request):
            self.assertRaises(
                requests.ConnectionError, http_client.client_request,
                TestClient(http_client), "GET", "/resource")

    def test_client_with_invalid_service_catalog(self):
        http_client = client.HTTPClient(FakeAuthPlugin())
        mock_request = mock.Mock()
        mock_request.side_effect = exceptions.EndpointException
        with mock.patch("requests.Session.request", mock_request):
            self.assertRaises(
                exceptions.EndpointException, http_client.client_request,
                TestClient(http_client), "GET", "/resource")

    def test_client_with_connection_refused(self):
        http_client = client.HTTPClient(FakeAuthPlugin())
        mock_request = mock.Mock()
        mock_request.side_effect = exceptions.ConnectionRefused
        with mock.patch("requests.Session.request", mock_request):
            self.assertRaises(
                exceptions.ConnectionRefused, http_client.client_request,
                TestClient(http_client), "GET", "/resource")



# Copyright (c) 2015 Rackspace
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from unittest import mock


from solumclient.common import github
from solumclient.tests import base


class TestGitHubAuth(base.TestCase):
    fake_repo = "http://github.com/fakeuser/fakerepo.git"
    fake_trigger = "http://example.com/trigger/1"
    fake_username = 'fakeuser'
    fake_password = 'fakepassword'
    fake_token = 'faketoken'

    def test_invalid_repo(self):
        self.assertRaises(ValueError,
                          github.GitHubAuth,
                          "http://example.com")

    def test_auth_header_username_password(self):
        gha = github.GitHubAuth(self.fake_repo,
                                username=self.fake_username,
                                password=self.fake_password)
        # base64.b64encode('fakeuser:fakepassword') yields 'ZmFrZX...'
        expected_auth_header = {
            'Content-Type': 'application/json',
            'Authorization': 'Basic ZmFrZXVzZXI6ZmFrZXBhc3N3b3Jk',
        }
        self.assertEqual(expected_auth_header, gha.auth_header)

    @mock.patch('getpass.getpass')
    def test_auth_header_username_password_2fa(self, fake_getpass):
        gha = github.GitHubAuth(self.fake_repo,
                                username=self.fake_username,
                                password=self.fake_password)
        gha._otp_required = True
        fake_getpass.return_value = 'fakeonetime'
        expected_auth_header = {
            'Content-Type': 'application/json',
            'Authorization': 'Basic ZmFrZXVzZXI6ZmFrZXBhc3N3b3Jk',
            'x-github-otp': 'fakeonetime',
        }
        self.assertEqual(expected_auth_header, gha.auth_header)

    def test_auth_header_repo_token(self):
        gha = github.GitHubAuth(self.fake_repo,
                                repo_token=self.fake_token)
        expected_auth_header = {
            'Content-Type': 'application/json',
            'Authorization': 'token %s' % self.fake_token,
        }
        self.assertEqual(expected_auth_header, gha.auth_header)

    @mock.patch('httplib2.Http.request')
    def test_create_webhook(self, fake_request):
        gha = github.GitHubAuth(self.fake_repo,
                                repo_token=self.fake_token)
        fake_request.return_value = ({'status': '200'},
                                     '{"token": "%s"}' % self.fake_token)
        gha.create_repo_token = mock.MagicMock()
        gha.create_repo_token.return_value = 'token123'

        gha.create_webhook(self.fake_trigger)
        fake_request.assert_called_once_with(
            'https://api.github.com/repos/fakeuser/fakerepo/hooks',
            'POST',
            headers=mock.ANY,
            body=mock.ANY)

        expected_body = {
            "config": {
                "url": self.fake_trigger,
                "content_type": "json"},
            "name": "web",
            "events": ["pull_request", "commit_comment"]}

        actual_body = json.loads(fake_request.call_args[1]['body'])
        self.assertEqual(expected_body, actual_body)

    @mock.patch('httplib2.Http.request')
    def test_create_webhook_unittest_only(self, fake_request):
        gha = github.GitHubAuth(self.fake_repo,
                                username=self.fake_username,
                                password=self.fake_password)
        fake_request.return_value = ({'status': '200'},
                                     '{"token": "foo"}')

        gha.create_repo_token = mock.MagicMock()
        gha.create_repo_token.return_value = 'token123'

        gha.create_webhook(self.fake_trigger, workflow=['unittest'])
        fake_request.assert_called_once_with(
            'https://api.github.com/repos/fakeuser/fakerepo/hooks',
            'POST',
            headers=mock.ANY,
            body=mock.ANY)
        expected_body = {
            "config": {
                "url": self.fake_trigger + "?workflow=unittest",
                "content_type": "json"},
            "name": "web",
            "events": ["pull_request", "commit_comment"]}
        actual_body = json.loads(fake_request.call_args[1]['body'])
        self.assertEqual(expected_body, actual_body)

    @mock.patch('httplib2.Http.request')
    def test_create_webhook_unittest_build(self, fake_request):
        gha = github.GitHubAuth(self.fake_repo,
                                username=self.fake_username,
                                password=self.fake_password)
        fake_request.return_value = ({'status': '200'},
                                     '{"token": "foo"}')

        gha.create_repo_token = mock.MagicMock()
        gha.create_repo_token.return_value = 'token123'

        gha.create_webhook(self.fake_trigger, workflow=['unittest', 'build'])
        fake_request.assert_called_once_with(
            'https://api.github.com/repos/fakeuser/fakerepo/hooks',
            'POST',
            headers=mock.ANY,
            body=mock.ANY)
        expected_body = {
            "config": {
                "url": self.fake_trigger + "?workflow=unittest+build",
                "content_type": "json"},
            "name": "web",
            "events": ["pull_request", "commit_comment"]}
        actual_body = json.loads(fake_request.call_args[1]['body'])
        self.assertEqual(expected_body, actual_body)

    @mock.patch('httplib2.Http.request')
    def test_add_ssh_key(self, fake_request):
        gha = github.GitHubAuth(self.fake_repo,
                                username=self.fake_username,
                                password=self.fake_password)
        fake_request.return_value = ({'status': '200'},
                                     '{"token": "foo"}')
        fake_pub_key = 'foo'
        gha.add_ssh_key(public_key=fake_pub_key)
        fake_request.assert_called_once_with(
            'https://api.github.com/user/keys',
            'POST',
            headers=mock.ANY,
            body=mock.ANY)
        expected_body = {"key": "foo", "title": "devops@Solum"}
        actual_body = json.loads(fake_request.call_args[1]['body'])
        self.assertEqual(expected_body, actual_body)



# Copyright 2014 - Noorul Islam K M
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from solumclient.common.apiclient import client as api_client
from solumclient.common import auth
from solumclient.common import client

API_NAME = 'builder'
VERSION_MAP = {
    '1': 'solumclient.builder.v1.client.Client',
}


def Client(version, **kwargs):
    client_class = api_client.BaseClient.get_class(API_NAME, version,
                                                   VERSION_MAP)
    keystone_auth = auth.KeystoneAuthPlugin(
        username=kwargs.get('username'),
        password=kwargs.get('password'),
        tenant_name=kwargs.get('tenant_name'),
        token=kwargs.get('token'),
        auth_url=kwargs.get('auth_url'),
        endpoint=kwargs.get('endpoint'))
    http_client = client.HTTPClient(keystone_auth)
    return client_class(http_client)


def get_client(api_version, **kwargs):
    """Get an authtenticated client.

    This is based on the credentials in the keyword args.

    :param api_version: the API version to use
    :param kwargs: keyword args containing credentials, either:
            * os_auth_token: pre-existing token to re-use
            * endpoint: solum API endpoint
            or:
            * os_username: name of user
            * os_password: user's password
            * os_auth_url: endpoint to authenticate against
            * os_tenant_name: name of tenant
    """
    cli_kwargs = {
        'username': kwargs.get('os_username'),
        'password': kwargs.get('os_password'),
        'tenant_name': kwargs.get('os_tenant_name'),
        'token': kwargs.get('os_auth_token'),
        'auth_url': kwargs.get('os_auth_url'),
        'endpoint': kwargs.get('solum_url')
    }

    return Client(api_version, **cli_kwargs)



# Copyright 2013 - Noorul Islam K M
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from keystoneclient.v2_0 import client as ksclient
import testtools

from solumclient.common import auth
from solumclient.common import client
from solumclient.tests import base


@mock.patch.object(ksclient, 'Client')
class KeystoneAuthPluginTest(base.TestCase):
    def setUp(self):
        super(KeystoneAuthPluginTest, self).setUp()
        plugin = auth.KeystoneAuthPlugin(
            username="fake-username",
            password="fake-password",
            tenant_name="fake-tenant-name",
            project_domain_name="default",
            user_domain_name="default",
            auth_url="http://auth")
        self.cs = client.HTTPClient(auth_plugin=plugin)

    @testtools.skip("Skip it when found solution")
    def test_authenticate(self, mock_ksclient):
        self.cs.authenticate()
        mock_ksclient.assert_called_with(
            username="fake-username",
            password="fake-password",
            tenant_name="fake-tenant-name",
            project_domain_name="default",
            user_domain_name="default",
            auth_url="http://auth")

    def test_token_and_endpoint(self, mock_ksclient):
        plugin = auth.KeystoneAuthPlugin(
            endpoint="http://solum",
            token="test_token")
        self.cs = client.HTTPClient(auth_plugin=plugin)
        self.cs.authenticate()
        (token, endpoint) = self.cs.auth_plugin.token_and_endpoint(
            "fake-endpoint-type", "fake-service-type")
        self.assertEqual(token, "test_token")
        self.assertEqual("http://solum", endpoint)

    def test_token_and_endpoint_before_auth(self, mock_ksclient):
        (token, endpoint) = self.cs.auth_plugin.token_and_endpoint(
            "fake-endpoint-type", "fake-service-type")
        self.assertIsNone(token, None)
        self.assertIsNone(endpoint, None)

    @testtools.skip("Skip it when found solution")
    def test_endpoint_with_no_token(self, mock_ksclient):
        plugin = auth.KeystoneAuthPlugin(
            username="fake-username",
            password="fake-password",
            tenant_name="fake-tenant-name",
            project_domain_name="default",
            user_domain_name="default",
            auth_url="http://auth",
            endpoint="http://solum")
        self.cs = client.HTTPClient(auth_plugin=plugin)
        self.cs.authenticate()
        mock_ksclient.assert_called_with(
            username="fake-username",
            password="fake-password",
            tenant_name="fake-tenant-name",
            auth_url="http://auth")
        (token, endpoint) = self.cs.auth_plugin.token_and_endpoint(
            "fake-endpoint-type", "fake-service-type")
        self.assertIsInstance(token, mock.MagicMock)
        self.assertEqual("http://solum", endpoint)


@mock.patch.object(ksclient, 'Client')
class KeystoneAuthPluginTokenTest(base.TestCase):
    def test_token_and_endpoint(self, mock_ksclient):
        plugin = auth.KeystoneAuthPlugin(
            token="fake-token",
            endpoint="http://solum")
        cs = client.HTTPClient(auth_plugin=plugin)

        cs.authenticate()
        (token, endpoint) = cs.auth_plugin.token_and_endpoint(
            "fake-endpoint-type", "fake-service-type")
        self.assertEqual('fake-token', token)
        self.assertEqual('http://solum', endpoint)



# Copyright 2013 - Noorul Islam K M
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from solumclient.common.apiclient import base as apiclient_base
from solumclient.common.apiclient import exceptions
from solumclient.common import base as solum_base
from solumclient.common import exc
from solumclient.common import yamlutils

from oslo_utils import uuidutils


class Requirement(apiclient_base.Resource):
    def __repr__(self):
        return "<Requirement %s>" % self._info


class ServiceReference(apiclient_base.Resource):
    def __repr__(self):
        return "<ServiceReference %s>" % self._info


class Artifact(apiclient_base.Resource):
    def __repr__(self):
        return "<Artifact %s>" % self._info

    def _add_requirements_details(self, req_list):
        return [Requirement(None, res, loaded=True)
                for res in req_list if req_list]

    def _add_details(self, info):
        for (k, v) in info.items():
            try:
                if k == 'requirements':
                    v = self._add_requirements_details(v)
                setattr(self, k, v)
                self._info[k] = v
            except AttributeError:
                # In this case we already defined the attribute on the class
                pass


class Plan(apiclient_base.Resource):
    def __repr__(self):
        return "<Plan %s>" % self._info

    def _add_artifact_details(self, artf_list):
        return [Artifact(None, res, loaded=True)
                for res in artf_list if artf_list]

    def _add_services_details(self, serv_list):
        return [ServiceReference(None, res, loaded=True)
                for res in serv_list if serv_list]

    def _add_details(self, info):
        for (k, v) in info.items():
            try:
                if k == 'artifacts':
                    v = self._add_artifact_details(v)
                elif k == 'services':
                    v = self._add_services_details(v)
                setattr(self, k, v)
                self._info[k] = v
            except AttributeError:
                # In this case we already defined the attribute on the class
                pass


class PlanManager(solum_base.CrudManager, solum_base.FindMixin):
    resource_class = Plan
    collection_key = 'plans'
    key = 'plan'

    def list(self, **kwargs):
        kwargs = self._filter_kwargs(kwargs)
        kwargs.setdefault("headers", kwargs.get("headers", {}))
        kwargs['headers']['Content-Type'] = 'x-application/yaml'
        resp = self.client.get(
            self.build_url(base_url="/v1", **kwargs), **kwargs)
        try:
            resp_plan = yamlutils.load(resp.content)
        except ValueError as e:
            raise exc.CommandException(message='Could not load Plan. '
                                       'Reason: %s' % e.message)
        return [Plan(self, res, loaded=True) for res in resp_plan if res]

    def create(self, plan, **kwargs):
        kwargs = self._filter_kwargs(kwargs)
        kwargs['data'] = plan
        kwargs.setdefault("headers", kwargs.get("headers", {}))
        kwargs['headers']['Content-Type'] = 'x-application/yaml'
        try:
            resp = self.client.post(
                self.build_url(base_url="/v1", **kwargs), **kwargs)
        except Exception as e:
            message = vars(e).get('details', str(e))
            raise exceptions.BadRequest(message=message)
        try:
            resp_plan = yamlutils.load(resp.content)
        except ValueError as e:
            raise exc.CommandException(message='Could not load Plan. '
                                       'Reason: %s' % e.message)
        return Plan(self, resp_plan)

    def _get(self, url, response_key=None):
        kwargs = {'headers': {}}
        kwargs['headers']['Content-Type'] = 'x-application/yaml'
        resp = self.client.get(url, **kwargs)
        try:
            resp_plan = yamlutils.load(resp.content)
        except ValueError as e:
            raise exc.CommandException(message='Could not load Plan. '
                                       'Reason: %s' % e.message)

        return Plan(self, resp_plan, loaded=True)

    def get(self, **kwargs):
        return super(PlanManager, self).get(base_url="/v1", **kwargs)

    def find(self, **kwargs):
        if 'plan_id' in kwargs:
            return super(PlanManager, self).get(base_url="/v1", **kwargs)
        elif 'name_or_id' in kwargs:
            name_or_uuid = kwargs['name_or_id']
            try:
                if uuidutils.is_uuid_like(name_or_uuid):
                    return super(PlanManager, self).get(base_url="/v1",
                                                        plan_id=name_or_uuid)
                else:
                    return super(PlanManager, self).findone(name=name_or_uuid)
            except exceptions.NoUniqueMatch:
                raise exc.NotUnique(resource='Plan')

    def update(self, plan, **kwargs):
        kwargs = self._filter_kwargs(kwargs)
        kwargs['data'] = plan
        kwargs.setdefault("headers", kwargs.get("headers", {}))
        kwargs['headers']['Content-Type'] = 'x-application/yaml'
        resp = self.client.put(self.build_url(base_url="/v1", **kwargs),
                               **kwargs)
        try:
            resp_plan = yamlutils.load(resp.content)
        except ValueError as e:
            raise exc.CommandException(message='Could not load Plan. '
                                       'Reason: %s' % e.message)
        return self.resource_class(self, resp_plan)

    def delete(self, **kwargs):
        return super(PlanManager, self).delete(base_url="/v1", **kwargs)



# Copyright 2014 - Noorul Islam K M
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from solumclient.builder.v1 import client as builder_client
from solumclient.builder.v1 import image
from solumclient.common.apiclient import fake_client
from solumclient.tests import base

image_fixture = {
    'uri': 'http://example.com/v1/images/i1',
    'name': 'php-web-app',
    'source_uri': 'git://example.com/project/app.git',
    'type': 'image',
    'description': 'A php web application',
    'tags': ['small'],
    'project_id': '1dae5a09ef2b4d8cbf3594b0eb4f6b94',
    'user_id': '55f41cf46df74320b9486a35f5d28a11',
}

fixtures_get = {
    '/v1/images/i1': {
        'GET': (
            {},
            image_fixture
        ),
    }
}


fixtures_create = {
    '/v1/images': {
        'POST': (
            {},
            image_fixture
        ),
    }
}


class ImageManagerTest(base.TestCase):

    def test_create(self):
        fake_http_client = fake_client.FakeHTTPClient(fixtures=fixtures_create)
        api_client = builder_client.Client(fake_http_client)
        mgr = image.ImageManager(api_client)
        image_obj = mgr.create()
        self.assertIn('Image', repr(image_obj))
        self.assertEqual(image_fixture['uri'], image_obj.uri)
        self.assertEqual(image_fixture['type'], image_obj.type)
        self.assertEqual(image_fixture['project_id'], image_obj.project_id)
        self.assertEqual(image_fixture['user_id'], image_obj.user_id)

    def test_get(self):
        fake_http_client = fake_client.FakeHTTPClient(fixtures=fixtures_get)
        api_client = builder_client.Client(fake_http_client)
        mgr = image.ImageManager(api_client)
        image_obj = mgr.get(image_id='i1')
        self.assertIn('Image', repr(image_obj))
        self.assertEqual(image_fixture['uri'], image_obj.uri)
        self.assertEqual(image_fixture['type'], image_obj.type)
        self.assertEqual(image_fixture['project_id'], image_obj.project_id)
        self.assertEqual(image_fixture['user_id'], image_obj.user_id)



# Copyright 2015 - Rackspace
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from solumclient.common.apiclient import base as apiclient_base
from solumclient.common.apiclient import exceptions
from solumclient.common import base as solum_base
from solumclient.common import exc

from oslo_utils import uuidutils


class Workflow(apiclient_base.Resource):
    def __repr__(self):
        return "<Workflow %s>" % self._info


class UserLog(apiclient_base.Resource):
    def __repr__(self):
        return "<Log %s>" % self._info


class WorkflowManager(solum_base.CrudManager, solum_base.FindMixin):
    resource_class = Workflow
    collection_key = 'workflows'
    key = 'workflow'

    def list(self, **kwargs):
        self.app_id = kwargs.pop('app_id')
        self.base_url = '/v1/apps/%s' % self.app_id
        return (super(WorkflowManager, self).list(
                base_url=self.base_url, **kwargs))

    def create(self, **kwargs):
        self.app_id = kwargs.get('app_id')
        self.base_url = '/v1/apps/%s' % self.app_id
        return (super(WorkflowManager, self).create(
                base_url=self.base_url, **kwargs))

    def get(self, **kwargs):
        self.app_id = kwargs.pop('app_id')
        self.base_url = '/v1/apps/%s' % self.app_id
        return (super(WorkflowManager, self).get(
                base_url=self.base_url, **kwargs))

    def logs(self, **kwargs):
        self.app_id = kwargs.get('app_id')
        self.base_url = '/v1/apps/%s' % self.app_id
        self.resource_class = UserLog
        url = self.build_url(self.base_url, **kwargs)
        rev_or_uuid = kwargs['revision_or_id']
        try:
            if uuidutils.is_uuid_like(rev_or_uuid):
                workflow_id = rev_or_uuid
            else:
                wf = self.find(**kwargs)
                workflow_id = wf.id
        except exceptions.NoUniqueMatch:
            raise exc.NotUnique(resource='Workflow')

        url += '/%s/logs/' % workflow_id
        return self._list(url)

    def find(self, **kwargs):
        self.app_id = kwargs.get('app_id')
        self.base_url = '/v1/apps/%s' % self.app_id
        if 'workflow_id' in kwargs:
            return (super(WorkflowManager, self).get(
                    base_url=self.base_url, **kwargs))
        elif 'revision_or_id' in kwargs:
            rev_or_uuid = kwargs['revision_or_id']
            try:
                if uuidutils.is_uuid_like(rev_or_uuid):
                    return super(WorkflowManager, self).get(
                        base_url=self.base_url,
                        workflow_id=rev_or_uuid)
                else:
                    return super(WorkflowManager, self).findone(
                        app_id=self.app_id, wf_id=rev_or_uuid)
            except exceptions.NoUniqueMatch:
                raise exc.NotUnique(resource='Workflow')



# Copyright 2014 - Rackspace
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from solumclient.builder.v1 import image
from solumclient.common.apiclient import client
from solumclient.common.apiclient import fake_client
from solumclient.tests import base
from solumclient.v1 import languagepack

languagepack_list = [
    {
        'uri': 'http://example.com/v1/language_packs/x1',
        'name': 'database',
        'language_pack_type': 'python',
        'description': 'Python Language pack',
        'tags': ['python'],
        'project_id': '1dae5a09ef2b4d8cbf3594b0eb4f6b94',
        'user_id': '55f41cf46df74320b9486a35f5d28a11',
        'language_implementation': 'python',
        'compiler_versions': '2.6',
        'os_platform': 'ubuntu 12.04',
        'attr_blob': '',
        'service_id': 1
    },
    {
        'uri': 'http://example.com/v1/language_packs/x2',
        'name': 'database',
        'language_pack_type': 'java',
        'description': 'Java Language pack',
        'tags': ['java'],
        'project_id': '1dae5a09ef2b4d8cbf3594b0eb4f6b94',
        'user_id': '55f41cf46df74320b9486a35f5d28a11',
        'language_implementation': 'java',
        'compiler_versions': '7.0',
        'os_platform': 'ubuntu 12.04',
        'attr_blob': '',
        'service_id': 1
    }
]

languagepack_fixture = {
    'uri': 'http://example.com/v1/language_packs/x1',
    'name': 'database',
    'language_pack_type': 'java',
    'description': 'Java Language pack',
    'tags': ['java'],
    'project_id': '1dae5a09ef2b4d8cbf3594b0eb4f6b94',
    'user_id': '55f41cf46df74320b9486a35f5d28a11',
    'language_implementation': 'java',
    'compiler_versions': '7.0',
    'os_platform': 'ubuntu 12.04',
    'attr_blob': '',
    'service_id': 1
}

fixtures_list = {
    '/v1/language_packs': {
        'GET': (
            {},
            languagepack_list
        ),
    }
}


fixtures_get = {
    '/v1/language_packs/x1': {
        'GET': (
            {},
            languagepack_fixture
        ),
    }
}


fixtures_create = {
    '/v1/language_packs': {
        'POST': (
            {},
            languagepack_fixture
        ),
    }
}

image_fixture = {
    'name': 'lp1',
    'source_uri': 'github.com/test',
    'lp_metadata': 'sample_lp_metadata'
}

fixtures_build = {
    '/v1/images': {
        'POST': (
            {},
            image_fixture
        ),
    }
}


class LanguagePackManagerTest(base.TestCase):

    def assert_lp_object(self, lp_obj):
        self.assertIn('LanguagePack', repr(lp_obj))
        self.assertEqual(languagepack_fixture['uri'], lp_obj.uri)
        self.assertEqual(languagepack_fixture['language_pack_type'],
                         lp_obj.language_pack_type)
        self.assertEqual(languagepack_fixture['project_id'], lp_obj.project_id)
        self.assertEqual(languagepack_fixture['user_id'], lp_obj.user_id)

    def test_list_all(self):
        fake_http_client = fake_client.FakeHTTPClient(fixtures=fixtures_list)
        api_client = client.BaseClient(fake_http_client)
        mgr = languagepack.LanguagePackManager(api_client)
        languagepacks = mgr.list()
        self.assertEqual(2, len(languagepacks))
        self.assertIn('LanguagePack', repr(languagepacks[0]))
        self.assertEqual(languagepack_list[0]['uri'],
                         languagepacks[0].uri)
        self.assertEqual(languagepack_list[1]['uri'],
                         languagepacks[1].uri)

    def test_create(self):
        fake_http_client = fake_client.FakeHTTPClient(fixtures=fixtures_create)
        api_client = client.BaseClient(fake_http_client)
        mgr = languagepack.LanguagePackManager(api_client)
        languagepack_obj = mgr.create()
        self.assert_lp_object(languagepack_obj)

    def test_get(self):
        fake_http_client = fake_client.FakeHTTPClient(fixtures=fixtures_get)
        api_client = client.BaseClient(fake_http_client)
        mgr = languagepack.LanguagePackManager(api_client)
        languagepack_obj = mgr.get(lp_id='x1')
        self.assert_lp_object(languagepack_obj)

    def test_build(self):
        fake_http_client = fake_client.FakeHTTPClient(fixtures=fixtures_build)
        api_client = client.BaseClient(fake_http_client)
        mgr = image.ImageManager(api_client)
        image_obj = mgr.create(name='lp1',
                               source_uri='github.com/test',
                               lp_metadata='sample_lp_metadata')
        self.assert_image_object(image_obj)

    def assert_image_object(self, image_obj):
        self.assertIn('Image', repr(image_obj))
        self.assertEqual(image_fixture['source_uri'], image_obj.source_uri)
        self.assertEqual(image_fixture['name'], image_obj.name)
        self.assertEqual(image_fixture['lp_metadata'], image_obj.lp_metadata)
